from typing import List

import numpy as np

from hats.pixel_math.healpix_pixel import HealpixPixel


def get_pixel_argsort(pixels: List[HealpixPixel]):
    """Perform an indirect sort of a list of HealpixPixels.

    This will order the pixels according to a breadth-first traversal of the healpix
    pixel hierarchy, not merely by increasing order by Norder/Npix (depth-first ordering).
    This is similar to ordering fully by _healpix_29.

    Args:
        pixels (List[HealpixPixel]): array to sort

    Returns:
        array of indices that sort the pixels in breadth-first order.
    """
    if pixels is None or len(pixels) == 0:
        return []
    # Construct a parallel list of exploded, high order pixels.
    highest_order = np.max(pixels).order

    constant_breadth_pixel = [pixel.pixel * (4 ** (highest_order - pixel.order)) for pixel in pixels]

    # Get the argsort of the higher order array.
    return np.argsort(constant_breadth_pixel, kind="stable")


def sort_pixels(pixels: List[HealpixPixel]):
    """Perform an sort of a list of HealpixPixels.

    This will order the pixels according to a breadth-first traversal of the healpix
    pixel hierarchy, not merely by increasing order by Norder/Npix (depth-first ordering).
    This is similar to ordering fully by _healpix_29.

    Args:
        pixels (List[HealpixPixel]): array to sort

    Returns:
        pixels sorted in breadth-first order
    """
    if pixels is None or len(pixels) == 0:
        return []
    argsort = get_pixel_argsort(pixels)
    return np.array(pixels)[argsort]


def get_pixels_from_intervals(intervals: np.ndarray, tree_order: int) -> np.ndarray:
    """Computes an array of HEALPix [order, pixel] for an array of intervals

    Args:
        intervals (np.ndarray): Array of intervals of the start and end pixel numbers of HEALPix pixels.
            Must be NESTED numbering scheme.
        tree_order (int): The order of the pixel numbers in the interval array

    Returns (np.ndarray):
        An array of [order, pixel] in NESTED numbering scheme for each interval in the array.
    """
    if intervals.shape[0] == 0:
        return np.empty((0, 2), dtype=np.int64)
    orders = np.full(intervals.shape[0], fill_value=-1)
    pixels = np.full(intervals.shape[0], fill_value=-1)
    # alignment uses (-1, -1) as a missing pixel, so we can't use the HEALPix math on these elements
    non_negative_mask = intervals.T[0] >= 0
    start_intervals = intervals.T[0][non_negative_mask]
    end_intervals = intervals.T[1][non_negative_mask]
    orders[non_negative_mask] = tree_order - (np.int64(np.log2(end_intervals - start_intervals)) >> 1)
    pixels[non_negative_mask] = start_intervals >> 2 * (tree_order - orders[non_negative_mask])
    return np.array([orders, pixels]).T



import re
from pathlib import Path
from typing import Iterable, List, Optional, Union

from jproperties import Properties
from pydantic import BaseModel, ConfigDict, Field, field_serializer, field_validator, model_validator
from typing_extensions import Self
from upath import UPath

from hats.catalog.catalog_type import CatalogType
from hats.io import file_io

## catalog_name, catalog_type, and total_rows are allowed for ALL types
CATALOG_TYPE_ALLOWED_FIELDS = {
    CatalogType.OBJECT: ["ra_column", "dec_column", "default_columns"],
    CatalogType.SOURCE: ["primary_catalog", "ra_column", "dec_column", "default_columns"],
    CatalogType.ASSOCIATION: [
        "primary_catalog",
        "primary_column",
        "primary_column_association",
        "join_catalog",
        "join_column",
        "join_column_association",
        "contains_leaf_files",
    ],
    CatalogType.INDEX: ["primary_catalog", "indexing_column", "extra_columns"],
    CatalogType.MARGIN: ["primary_catalog", "margin_threshold", "ra_column", "dec_column", "default_columns"],
}

## catalog_name, catalog_type, and total_rows are required for ALL types
CATALOG_TYPE_REQUIRED_FIELDS = {
    CatalogType.OBJECT: ["ra_column", "dec_column"],
    CatalogType.SOURCE: ["ra_column", "dec_column"],
    CatalogType.ASSOCIATION: [
        "primary_catalog",
        "primary_column",
        "join_catalog",
        "join_column",
        "contains_leaf_files",
    ],
    CatalogType.INDEX: ["primary_catalog", "indexing_column"],
    CatalogType.MARGIN: ["primary_catalog", "margin_threshold"],
}

# All additional properties in the HATS recommendation.
EXTRA_ALLOWED_FIELDS = [
    "addendum_did",
    "bib_reference",
    "bib_reference_url",
    "creator_did",
    "data_ucd",
    "hats_builder",
    "hats_cols_sort",
    "hats_cols_survey_id",
    "hats_coordinate_epoch",
    "hats_copyright",
    "hats_creation_date",
    "hats_creator",
    "hats_estsize",
    "hats_frame",
    "hats_max_rows",
    "hats_order",
    "hats_progenitor_url",
    "hats_release_date",
    "hats_service_url",
    "hats_status",
    "hats_version",
    "moc_sky_fraction",
    "obs_ack",
    "obs_copyright",
    "obs_copyright_url",
    "obs_description",
    "obs_regime",
    "obs_title",
    "prov_progenitor",
    "publisher_id",
    "t_max",
    "t_min",
]


class TableProperties(BaseModel):
    """Container class for catalog metadata"""

    catalog_name: str = Field(alias="obs_collection")
    catalog_type: CatalogType = Field(alias="dataproduct_type")
    total_rows: int = Field(alias="hats_nrows")

    ra_column: Optional[str] = Field(default=None, alias="hats_col_ra")
    dec_column: Optional[str] = Field(default=None, alias="hats_col_dec")
    default_columns: Optional[List[str]] = Field(default=None, alias="hats_cols_default")
    """Which columns should be read from parquet files, when user doesn't otherwise specify."""

    primary_catalog: Optional[str] = Field(default=None, alias="hats_primary_table_url")
    """Reference to object catalog. Relevant for nested, margin, association, and index."""

    margin_threshold: Optional[float] = Field(default=None, alias="hats_margin_threshold")
    """Threshold of the pixel boundary, expressed in arcseconds."""

    primary_column: Optional[str] = Field(default=None, alias="hats_col_assn_primary")
    """Column name in the primary (left) side of join."""

    primary_column_association: Optional[str] = Field(default=None, alias="hats_col_assn_primary_assn")
    """Column name in the association table that matches the primary (left) side of join."""

    join_catalog: Optional[str] = Field(default=None, alias="hats_assn_join_table_url")
    """Catalog name for the joining (right) side of association."""

    join_column: Optional[str] = Field(default=None, alias="hats_col_assn_join")
    """Column name in the joining (right) side of join."""

    join_column_association: Optional[str] = Field(default=None, alias="hats_col_assn_join_assn")
    """Column name in the association table that matches the joining (right) side of join."""

    contains_leaf_files: Optional[bool] = Field(default=None, alias="hats_assn_leaf_files")
    """Whether or not the association catalog contains leaf parquet files."""

    indexing_column: Optional[str] = Field(default=None, alias="hats_index_column")
    """Column that we provide an index over."""

    extra_columns: Optional[List[str]] = Field(default=None, alias="hats_index_extra_column")
    """Any additional payload columns included in index."""

    ## Allow any extra keyword args to be stored on the properties object.
    model_config = ConfigDict(extra="allow", populate_by_name=True, use_enum_values=True)

    @field_validator("default_columns", "extra_columns", mode="before")
    @classmethod
    def space_delimited_list(cls, str_value: str) -> List[str]:
        """Convert a space-delimited list string into a python list of strings."""
        if isinstance(str_value, str):
            # Split on a few kinds of delimiters (just to be safe), and remove duplicates
            return list(filter(None, re.split(";| |,|\n", str_value)))
        ## Convert empty strings and empty lists to None
        return str_value if str_value else None

    @field_serializer("default_columns", "extra_columns")
    def serialize_as_space_delimited_list(self, str_list: Iterable[str]) -> str:
        """Convert a python list of strings into a space-delimited string."""
        if str_list is None or len(str_list) == 0:
            return None
        return " ".join(str_list)

    @model_validator(mode="after")
    def check_allowed_and_required(self) -> Self:
        """Check that type-specific fields are appropriate, and required fields are set."""
        explicit_keys = set(
            self.model_dump(by_alias=False, exclude_none=True).keys() - self.__pydantic_extra__.keys()
        )

        allowed_keys = set(
            CATALOG_TYPE_ALLOWED_FIELDS[self.catalog_type] + ["catalog_name", "catalog_type", "total_rows"]
        )
        non_allowed = explicit_keys - allowed_keys
        if len(non_allowed) > 0:
            raise ValueError(f"Unexpected property for table type {self.catalog_type}: {non_allowed}")

        required_keys = set(
            CATALOG_TYPE_REQUIRED_FIELDS[self.catalog_type] + ["catalog_name", "catalog_type", "total_rows"]
        )
        missing_required = required_keys - explicit_keys
        if len(missing_required) > 0:
            raise ValueError(
                f"Missing required property for table type {self.catalog_type}: {missing_required}"
            )

        # Check against all known properties - catches typos.
        non_allowed = set(self.__pydantic_extra__.keys()) - set(EXTRA_ALLOWED_FIELDS)
        if len(non_allowed) > 0:
            raise ValueError(f"Unexpected extra property: {non_allowed}")
        return self

    def copy_and_update(self, **kwargs):
        """Create a validated copy of these table properties, updating the fields provided in kwargs."""
        new_properties = self.model_copy(update=kwargs)
        TableProperties.model_validate(new_properties)
        return new_properties

    def explicit_dict(self):
        """Create a dict, based on fields that have been explicitly set, and are not "extra" keys."""
        explicit = self.model_dump(by_alias=False, exclude_none=True)
        extra_keys = self.__pydantic_extra__.keys()
        return {key: val for key, val in explicit.items() if key not in extra_keys}

    def __str__(self):
        """Friendly string representation based on named fields."""
        parameters = self.explicit_dict()
        formatted_string = ""
        for name, value in parameters.items():
            formatted_string += f"  {name} {value}\n"
        return formatted_string

    @classmethod
    def read_from_dir(cls, catalog_dir: Union[str, Path, UPath]) -> Self:
        """Read field values from a java-style properties file."""
        file_path = file_io.get_upath(catalog_dir) / "properties"
        if not file_io.does_file_or_directory_exist(file_path):
            raise FileNotFoundError(f"No properties file found where expected: {str(file_path)}")
        p = Properties()
        with file_path.open("rb") as f:
            p.load(f, "utf-8")
        return cls(**p.properties)

    def to_properties_file(self, catalog_dir: Union[str, Path, UPath]) -> Self:
        """Write fields to a java-style properties file."""
        # pylint: disable=protected-access
        parameters = self.model_dump(by_alias=True, exclude_none=True)
        properties = Properties()
        properties.properties = parameters
        properties._key_order = parameters.keys()
        file_path = file_io.get_upath(catalog_dir) / "properties"
        with file_path.open("wb") as _file:
            properties.store(_file, encoding="utf-8", initial_comments="HATS catalog", timestamp=False)



import pytest

from hats.pixel_math import HealpixPixel
from hats.pixel_tree.pixel_tree import PixelTree


def assert_pixel_tree_has_nodes_in_catalog(tree, catalog):
    """assert tree contains the same nodes as the catalog"""
    for pixel in catalog.get_healpix_pixels():
        assert tree.contains((pixel.order, pixel.pixel))


def test_pixel_tree_small_sky(small_sky_catalog):
    """test pixel tree on small sky"""
    pixel_tree = PixelTree.from_healpix(small_sky_catalog.get_healpix_pixels())
    assert len(pixel_tree) == len(small_sky_catalog.get_healpix_pixels())
    assert_pixel_tree_has_nodes_in_catalog(pixel_tree, small_sky_catalog)


def test_pixel_tree_small_sky_order1(small_sky_order1_catalog):
    """test pixel tree on small sky order1"""
    pixel_tree = PixelTree.from_healpix(small_sky_order1_catalog.get_healpix_pixels())
    assert_pixel_tree_has_nodes_in_catalog(pixel_tree, small_sky_order1_catalog)


def test_pixel_tree_small_sky_from_list(small_sky_catalog, small_sky_pixels):
    """test pixel tree on small sky"""
    pixel_tree = PixelTree.from_healpix(small_sky_pixels)
    assert len(pixel_tree) == len(small_sky_pixels)
    assert_pixel_tree_has_nodes_in_catalog(pixel_tree, small_sky_catalog)


def test_pixel_tree_small_sky_order1_from_list(small_sky_order1_catalog, small_sky_order1_pixels):
    """test pixel tree on small sky order1"""
    pixel_tree = PixelTree.from_healpix(small_sky_order1_pixels)
    assert_pixel_tree_has_nodes_in_catalog(pixel_tree, small_sky_order1_catalog)


def test_duplicate_pixel_raises_error():
    """test pixel tree raises error with duplicate pixels"""
    partition_info = [
        HealpixPixel(0, 11),
    ]
    PixelTree.from_healpix(partition_info)
    info_with_duplicate = [
        HealpixPixel(0, 11),
        HealpixPixel(0, 11),
    ]
    with pytest.raises(ValueError):
        PixelTree.from_healpix(info_with_duplicate)


def test_pixel_duplicated_at_different_order_raises_error():
    """test pixel tree raises error with duplicate pixels at different orders"""
    info_with_duplicate = [
        HealpixPixel(0, 11),
        HealpixPixel(1, 44),  # overlaps with (0,11)
    ]
    with pytest.raises(ValueError):
        PixelTree.from_healpix(info_with_duplicate)



from typing import List

import numpy as np
import pyarrow.compute as pc
import pyarrow.dataset as pds

from hats.catalog.dataset import Dataset
from hats.io import paths
from hats.pixel_math import HealpixPixel
from hats.pixel_math.healpix_pixel_function import get_pixel_argsort


class IndexCatalog(Dataset):
    """An index into HATS Catalog for enabling fast lookups on non-spatial values.

    Note that this is not a true "HATS Catalog", as it is not partitioned spatially.
    """

    def loc_partitions(self, ids) -> List[HealpixPixel]:
        """Find the set of partitions in the primary catalog for the ids provided.

        Args:
            ids: the values of the indexing column (e.g. 87,543)
        Returns:
            partitions of leaf parquet files in the primary catalog
            that may contain rows for the id values
        """
        metadata_file = paths.get_parquet_metadata_pointer(self.catalog_base_dir)
        dataset = pds.parquet_dataset(metadata_file, filesystem=metadata_file.fs)

        # There's a lot happening in a few pyarrow dataset methods:
        # We create a simple pyarrow expression that roughly corresponds to a SQL statement like
        #   WHERE id_column IN (<ids>)
        # We stay in pyarrow to group by Norder/Npix to aggregate the results unique values.
        # After that convert into pandas, as this handles the integer type conversions
        # (uint8 and uint64 aren't always friendly between pyarrow and the rest of python),
        # and offers easy iteration to create our HealpixPixel list.
        filtered = dataset.filter(pc.field(self.catalog_info.indexing_column).isin(ids)).to_table()
        unique_pixel_dataframe = filtered.group_by(["Norder", "Npix"]).aggregate([]).to_pandas()

        loc_partitions = [
            HealpixPixel(order, pixel)
            for order, pixel in zip(
                unique_pixel_dataframe["Norder"],
                unique_pixel_dataframe["Npix"],
            )
        ]
        # Put the partitions in stable order (by nested healpix ordering).
        argsort = get_pixel_argsort(loc_partitions)
        loc_partitions = np.array(loc_partitions)[argsort]

        return loc_partitions



"""Utilities for performing fun math on healpix pixels"""

from .healpix_pixel import HealpixPixel
from .healpix_pixel_convertor import HealpixInputTypes, get_healpix_pixel
from .margin_bounding import check_margin_bounds
from .partition_stats import empty_histogram, generate_alignment, generate_histogram
from .pixel_margins import get_margin
from .spatial_index import compute_spatial_index, spatial_index_to_healpix



from pathlib import Path

from hats.io.file_io import (
    append_paths_to_pointer,
    directory_has_contents,
    does_file_or_directory_exist,
    find_files_matching_path,
    get_directory_contents,
    is_regular_file,
)


def test_file_or_dir_exist(small_sky_dir):
    assert does_file_or_directory_exist(small_sky_dir)

    catalog_info_string = small_sky_dir / "properties"
    assert does_file_or_directory_exist(catalog_info_string)


def test_file_or_dir_exist_false(small_sky_dir):
    assert not does_file_or_directory_exist(str(small_sky_dir) + "incorrect file")


def test_append_paths_to_pointer(tmp_path):
    test_paths = ["folder", "file.txt"]
    test_path = tmp_path / "folder" / "file.txt"
    assert str(append_paths_to_pointer(tmp_path, *test_paths)) == str(test_path)


def test_is_regular_file(small_sky_dir):
    properties_file = small_sky_dir / "properties"
    assert is_regular_file(properties_file)

    assert not is_regular_file(small_sky_dir)

    partition_dir = small_sky_dir / "Norder=0"
    assert not is_regular_file(partition_dir)


def test_find_files_matching_path(small_sky_dir):
    ## no_wildcard
    matching_files = find_files_matching_path(small_sky_dir, "properties")
    assert len(matching_files) == 1

    ## wilcard in the name (matches _metadata and _common_metadata)
    assert len(find_files_matching_path(small_sky_dir, "*metadata*")) == 2


def test_find_files_matching_path_directory(small_sky_order1_dir):
    assert len(find_files_matching_path(small_sky_order1_dir)) == 1

    ## wildcard in directory - will match all files at indicated depth
    assert len(find_files_matching_path(small_sky_order1_dir, "*", "*", "*", "*")) == 4


def test_directory_has_contents(small_sky_order1_dir, tmp_path):
    assert directory_has_contents(small_sky_order1_dir)
    assert not directory_has_contents(tmp_path)


def test_get_directory_contents(small_sky_order1_dir, tmp_path):
    small_sky_contents = get_directory_contents(small_sky_order1_dir)

    small_sky_paths = [Path(p) for p in small_sky_contents]

    expected = [
        "dataset",
        "partition_info.csv",
        "point_map.fits",
        "properties",
    ]

    expected = [small_sky_order1_dir / file_name for file_name in expected]

    assert small_sky_paths == expected

    assert len(get_directory_contents(tmp_path)) == 0



"""Tests of catalog functionality"""

import os

import astropy.units as u
import numpy as np
import pyarrow as pa
import pytest
from mocpy import MOC

import hats.pixel_math.healpix_shim as hp
from hats.catalog import Catalog, PartitionInfo, TableProperties
from hats.io import paths
from hats.io.file_io import read_fits_image
from hats.loaders import read_hats
from hats.pixel_math import HealpixPixel
from hats.pixel_math.box_filter import _generate_ra_strip_moc, generate_box_moc
from hats.pixel_math.validators import ValidatorsErrors
from hats.pixel_tree.pixel_tree import PixelTree


def test_catalog_load(catalog_info, catalog_pixels):
    catalog = Catalog(catalog_info, catalog_pixels)
    assert catalog.get_healpix_pixels() == catalog_pixels
    assert catalog.catalog_name == catalog_info.catalog_name

    for hp_pixel in catalog_pixels:
        assert hp_pixel in catalog.pixel_tree


def test_partition_info_pixel_input_types(catalog_info, catalog_pixels):
    partition_info = PartitionInfo.from_healpix(catalog_pixels)
    catalog = Catalog(catalog_info, partition_info)
    assert len(catalog.get_healpix_pixels()) == len(catalog_pixels)
    assert len(catalog.pixel_tree.get_healpix_pixels()) == len(catalog_pixels)
    for hp_pixel in catalog_pixels:
        assert hp_pixel in catalog.pixel_tree


def test_tree_pixel_input(catalog_info, catalog_pixels):
    tree = PixelTree.from_healpix(catalog_pixels)
    catalog = Catalog(catalog_info, tree)
    assert len(catalog.get_healpix_pixels()) == len(catalog_pixels)
    assert len(catalog.pixel_tree.get_healpix_pixels()) == len(catalog_pixels)
    for pixel in catalog_pixels:
        assert pixel in catalog.pixel_tree


def test_tree_pixel_input_list(catalog_info, catalog_pixels):
    catalog = Catalog(catalog_info, catalog_pixels)
    assert len(catalog.get_healpix_pixels()) == len(catalog_pixels)
    assert len(catalog.pixel_tree.get_healpix_pixels()) == len(catalog_pixels)
    for pixel in catalog_pixels:
        assert pixel in catalog.pixel_tree


def test_wrong_pixel_input_type(catalog_info):
    with pytest.raises(TypeError):
        Catalog(catalog_info, "test")
    with pytest.raises(TypeError):
        Catalog._get_pixel_tree_from_pixels("test")
    with pytest.raises(TypeError):
        Catalog._get_partition_info_from_pixels("test")


def test_get_pixels_list(catalog_info, catalog_pixels):
    catalog = Catalog(catalog_info, catalog_pixels)
    pixels = catalog.get_healpix_pixels()
    assert pixels == catalog_pixels


def test_load_catalog_small_sky(small_sky_dir, small_sky_schema):
    """Instantiate a catalog with 1 pixel"""
    cat = read_hats(small_sky_dir)

    assert isinstance(cat, Catalog)
    assert cat.catalog_name == "small_sky"
    assert len(cat.get_healpix_pixels()) == 1

    assert isinstance(cat.schema, pa.Schema)
    assert cat.schema.equals(small_sky_schema)


def test_load_catalog_small_sky_order1(small_sky_order1_dir):
    """Instantiate a catalog with 4 pixels"""
    cat = read_hats(small_sky_order1_dir)

    assert isinstance(cat, Catalog)
    assert cat.catalog_name == "small_sky_order1"
    assert len(cat.get_healpix_pixels()) == 4


def test_load_catalog_small_sky_order1_moc(small_sky_order1_dir):
    """Instantiate a catalog with 4 pixels"""
    cat = read_hats(small_sky_order1_dir)

    assert isinstance(cat, Catalog)
    assert cat.moc is not None
    counts_skymap = read_fits_image(paths.get_point_map_file_pointer(small_sky_order1_dir))
    skymap_order = hp.nside2order(hp.npix2nside(len(counts_skymap)))
    assert cat.moc.max_order == skymap_order
    assert np.all(cat.moc.flatten() == np.where(counts_skymap > 0))


def test_load_catalog_small_sky_source(small_sky_source_dir, small_sky_source_schema):
    """Instantiate a source catalog with 14 pixels"""
    cat = read_hats(small_sky_source_dir)

    assert isinstance(cat, Catalog)
    assert cat.catalog_name == "small_sky_source"
    assert len(cat.get_healpix_pixels()) == 14

    assert isinstance(cat.schema, pa.Schema)
    assert cat.schema.equals(small_sky_source_schema)


def test_max_coverage_order(small_sky_order1_catalog):
    assert small_sky_order1_catalog.get_max_coverage_order() >= small_sky_order1_catalog.moc.max_order
    assert (
        small_sky_order1_catalog.get_max_coverage_order()
        >= small_sky_order1_catalog.pixel_tree.get_max_depth()
    )
    high_moc_order = 8
    test_moc = MOC.from_depth29_ranges(
        max_depth=high_moc_order, ranges=small_sky_order1_catalog.moc.to_depth29_ranges
    )
    small_sky_order1_catalog.moc = test_moc
    assert small_sky_order1_catalog.get_max_coverage_order() == high_moc_order
    small_sky_order1_catalog.moc = None
    assert (
        small_sky_order1_catalog.get_max_coverage_order()
        == small_sky_order1_catalog.pixel_tree.get_max_depth()
    )


def test_cone_filter(small_sky_order1_catalog):
    ra = 315
    dec = -66.443
    radius = 0.1

    filtered_catalog = small_sky_order1_catalog.filter_by_cone(ra, dec, radius)
    filtered_pixels = filtered_catalog.get_healpix_pixels()

    assert len(filtered_pixels) == 1
    assert filtered_pixels == [HealpixPixel(1, 44)]

    assert (1, 44) in filtered_catalog.pixel_tree
    assert filtered_catalog.catalog_info.total_rows == 0

    assert filtered_catalog.moc is not None
    cone_moc = MOC.from_cone(
        lon=ra * u.deg,
        lat=dec * u.deg,
        radius=radius * u.arcsec,
        max_depth=small_sky_order1_catalog.get_max_coverage_order(),
    )
    assert filtered_catalog.moc == cone_moc.intersection(small_sky_order1_catalog.moc)


def test_cone_filter_big(small_sky_order1_catalog):
    filtered_catalog = small_sky_order1_catalog.filter_by_cone(315, -66.443, 30 * 3600)
    assert len(filtered_catalog.get_healpix_pixels()) == 4
    assert (1, 44) in filtered_catalog.pixel_tree
    assert (1, 45) in filtered_catalog.pixel_tree
    assert (1, 46) in filtered_catalog.pixel_tree
    assert (1, 47) in filtered_catalog.pixel_tree


@pytest.mark.timeout(5)
def test_cone_filter_multiple_order(catalog_info):
    catalog_pixel_list = [
        HealpixPixel(6, 30),
        HealpixPixel(7, 124),
        HealpixPixel(7, 5000),
    ]
    catalog = Catalog(catalog_info, catalog_pixel_list)
    filtered_catalog = catalog.filter_by_cone(47.1, 6, 30 * 3600)
    assert filtered_catalog.get_healpix_pixels() == [HealpixPixel(6, 30), HealpixPixel(7, 124)]


def test_cone_filter_empty(small_sky_order1_catalog):
    filtered_catalog = small_sky_order1_catalog.filter_by_cone(0, 0, 0.1)
    assert len(filtered_catalog.get_healpix_pixels()) == 0
    assert len(filtered_catalog.pixel_tree) == 0


def test_cone_filter_invalid_cone_center(small_sky_order1_catalog):
    with pytest.raises(ValueError, match=ValidatorsErrors.INVALID_DEC):
        small_sky_order1_catalog.filter_by_cone(0, -100, 0.1)
    with pytest.raises(ValueError, match=ValidatorsErrors.INVALID_DEC):
        small_sky_order1_catalog.filter_by_cone(0, 100, 0.1)
    with pytest.raises(ValueError, match=ValidatorsErrors.INVALID_RADIUS):
        small_sky_order1_catalog.filter_by_cone(0, 10, -1)


def test_polygonal_filter(small_sky_order1_catalog):
    polygon_vertices = [(282, -58), (282, -55), (272, -55), (272, -58)]
    filtered_catalog = small_sky_order1_catalog.filter_by_polygon(polygon_vertices)
    filtered_pixels = filtered_catalog.get_healpix_pixels()
    assert len(filtered_pixels) == 1
    assert filtered_pixels == [HealpixPixel(1, 46)]
    assert (1, 46) in filtered_catalog.pixel_tree
    assert filtered_catalog.catalog_info.total_rows == 0
    assert filtered_catalog.moc is not None
    ra, dec = np.array(polygon_vertices).T
    polygon_moc = MOC.from_polygon(
        lon=ra * u.deg,
        lat=dec * u.deg,
        max_depth=small_sky_order1_catalog.get_max_coverage_order(),
    )
    assert filtered_catalog.moc == polygon_moc.intersection(small_sky_order1_catalog.moc)


def test_polygonal_filter_with_cartesian_coordinates(small_sky_order1_catalog):
    sky_vertices = [(282, -58), (282, -55), (272, -55), (272, -58)]
    cartesian_vertices = hp.ang2vec(*np.array(sky_vertices).T, lonlat=True)
    filtered_catalog_1 = small_sky_order1_catalog.filter_by_polygon(sky_vertices)
    filtered_catalog_2 = small_sky_order1_catalog.filter_by_polygon(cartesian_vertices)
    assert filtered_catalog_1.get_healpix_pixels() == filtered_catalog_2.get_healpix_pixels()
    assert (1, 46) in filtered_catalog_1.pixel_tree
    assert (1, 46) in filtered_catalog_2.pixel_tree


def test_polygonal_filter_big(small_sky_order1_catalog):
    polygon_vertices = [(281, -69), (281, -25), (350, -25), (350, -69)]
    filtered_catalog = small_sky_order1_catalog.filter_by_polygon(polygon_vertices)
    assert len(filtered_catalog.get_healpix_pixels()) == 4
    assert (1, 44) in filtered_catalog.pixel_tree
    assert (1, 45) in filtered_catalog.pixel_tree
    assert (1, 46) in filtered_catalog.pixel_tree
    assert (1, 47) in filtered_catalog.pixel_tree


def test_polygonal_filter_multiple_order(catalog_info):
    catalog_pixel_list = [
        HealpixPixel(6, 30),
        HealpixPixel(7, 124),
        HealpixPixel(7, 5000),
    ]
    catalog = Catalog(catalog_info, catalog_pixel_list)
    polygon_vertices = [(47.1, 6), (64.5, 6), (64.5, 6.27), (47.1, 6.27)]
    filtered_catalog = catalog.filter_by_polygon(polygon_vertices)
    assert filtered_catalog.get_healpix_pixels() == [HealpixPixel(6, 30), HealpixPixel(7, 124)]


def test_polygonal_filter_empty(small_sky_order1_catalog):
    polygon_vertices = [(0, 0), (1, 1), (0, 2)]
    filtered_catalog = small_sky_order1_catalog.filter_by_polygon(polygon_vertices)
    assert len(filtered_catalog.get_healpix_pixels()) == 0
    assert len(filtered_catalog.pixel_tree) == 0


def test_polygonal_filter_invalid_coordinates(small_sky_order1_catalog):
    # Declination is over 90 degrees
    polygon_vertices = [(47.1, -100), (64.5, -100), (64.5, 6.27), (47.1, 6.27)]
    with pytest.raises(ValueError, match=ValidatorsErrors.INVALID_DEC):
        small_sky_order1_catalog.filter_by_polygon(polygon_vertices)
    # Right ascension should wrap, it does not throw an error
    polygon_vertices = [(470.1, 6), (470.5, 6), (64.5, 10.27), (47.1, 10.27)]
    small_sky_order1_catalog.filter_by_polygon(polygon_vertices)


def test_polygonal_filter_invalid_polygon(small_sky_order1_catalog):
    # The polygon must have a minimum of 3 vertices
    with pytest.raises(ValueError, match=ValidatorsErrors.INVALID_NUM_VERTICES):
        vertices = [(100.1, -20.3), (100.1, 40.3)]
        small_sky_order1_catalog.filter_by_polygon(vertices[:2])
    # The vertices should not have duplicates
    with pytest.raises(ValueError, match=ValidatorsErrors.DUPLICATE_VERTICES):
        vertices = [(100.1, -20.3), (100.1, -20.3), (280.1, -20.3), (280.1, 40.3)]
        small_sky_order1_catalog.filter_by_polygon(vertices)
    # The polygons should not be on a great circle
    with pytest.raises(ValueError, match=ValidatorsErrors.DEGENERATE_POLYGON):
        vertices = [(100.1, 40.3), (100.1, -20.3), (280.1, -20.3), (280.1, 40.3)]
        small_sky_order1_catalog.filter_by_polygon(vertices)
    with pytest.raises(ValueError, match=ValidatorsErrors.DEGENERATE_POLYGON):
        vertices = [(50.1, 0), (100.1, 0), (150.1, 0), (200.1, 0)]
        small_sky_order1_catalog.filter_by_polygon(vertices)


def test_box_filter_ra(small_sky_order1_catalog):
    ra = (280, 290)
    # The catalog pixels are distributed around the [270,0] degree range.
    filtered_catalog = small_sky_order1_catalog.filter_by_box(ra=ra)

    filtered_pixels = filtered_catalog.get_healpix_pixels()

    assert len(filtered_pixels) == 2
    assert filtered_pixels == [HealpixPixel(1, 44), HealpixPixel(1, 46)]

    assert (1, 44) in filtered_catalog.pixel_tree
    assert (1, 46) in filtered_catalog.pixel_tree
    assert len(filtered_catalog.pixel_tree.pixels[1]) == 2
    assert filtered_catalog.catalog_info.total_rows == 0

    assert filtered_catalog.moc is not None
    box_moc = generate_box_moc(ra=ra, dec=None, order=small_sky_order1_catalog.get_max_coverage_order())
    assert filtered_catalog.moc == box_moc.intersection(small_sky_order1_catalog.moc)


def test_box_filter_wrapped_ra(small_sky_order1_catalog):
    # The catalog pixels are distributed around the [270,0] degree range.
    filtered_catalog = small_sky_order1_catalog.filter_by_box(ra=(-10, 10))

    filtered_pixels = filtered_catalog.get_healpix_pixels()

    assert len(filtered_pixels) == 2
    assert filtered_pixels == [HealpixPixel(1, 44), HealpixPixel(1, 45)]

    assert (1, 44) in filtered_catalog.pixel_tree
    assert (1, 45) in filtered_catalog.pixel_tree
    assert len(filtered_catalog.pixel_tree.pixels[1]) == 2
    assert filtered_catalog.catalog_info.total_rows == 0


def test_box_filter_ra_divisions_edge_cases(small_sky_order1_catalog):
    # In this test we generate RA bands and their complements and compare the amount of
    # pixels from the catalog after filtering. We construct these complement regions in
    # a way that allows us to capture more pixels of the catalog. This is useful to test
    # that wide RA ranges (larger than 180 degrees) are correctly handled.

    # The catalog pixels are distributed around the [270,0] degree range.

    def assert_is_subset_of(catalog, catalog_complement):
        pixels_catalog = catalog.get_healpix_pixels()
        pixels_catalog_complement = catalog_complement.get_healpix_pixels()
        assert len(pixels_catalog) < len(pixels_catalog_complement)
        assert all(pixel in pixels_catalog_complement for pixel in pixels_catalog)

    filtered_catalog = small_sky_order1_catalog.filter_by_box(ra=(0, 180))
    filtered_catalog_complement = small_sky_order1_catalog.filter_by_box(ra=(180, 0))
    assert_is_subset_of(filtered_catalog, filtered_catalog_complement)

    filtered_catalog = small_sky_order1_catalog.filter_by_box(ra=(10, 50))
    filtered_catalog_complement = small_sky_order1_catalog.filter_by_box(ra=(50, 10))
    assert_is_subset_of(filtered_catalog, filtered_catalog_complement)

    filtered_catalog = small_sky_order1_catalog.filter_by_box(ra=(10, 220))
    filtered_catalog_complement = small_sky_order1_catalog.filter_by_box(ra=(220, 10))
    assert_is_subset_of(filtered_catalog, filtered_catalog_complement)

    filtered_catalog = small_sky_order1_catalog.filter_by_box(ra=(350, 200))
    filtered_catalog_complement = small_sky_order1_catalog.filter_by_box(ra=(200, 350))
    assert_is_subset_of(filtered_catalog, filtered_catalog_complement)

    filtered_catalog = small_sky_order1_catalog.filter_by_box(ra=(50, 200))
    filtered_catalog_complement = small_sky_order1_catalog.filter_by_box(ra=(200, 50))
    assert_is_subset_of(filtered_catalog, filtered_catalog_complement)


def test_box_filter_ra_pixel_tree_generation():
    """This method tests the pixel tree generation for the ra filter"""
    # The catalog pixels are distributed around the [270,0] degree range.
    moc = _generate_ra_strip_moc(ra_range=(0, 180), order=1)
    moc_complement = _generate_ra_strip_moc(ra_range=(180, 0), order=1)
    assert len(moc.flatten()) == len(moc_complement.flatten())

    moc = _generate_ra_strip_moc(ra_range=(10, 50), order=1)
    moc_complement = _generate_ra_strip_moc(ra_range=(50, 10), order=1)
    assert len(moc.flatten()) < len(moc_complement.flatten())

    moc = _generate_ra_strip_moc(ra_range=(10, 220), order=1)
    moc_complement = _generate_ra_strip_moc(ra_range=(220, 10), order=1)
    assert len(moc_complement.flatten()) < len(moc.flatten())

    moc = _generate_ra_strip_moc(ra_range=(200, 350), order=1)
    moc_complement = _generate_ra_strip_moc(ra_range=(350, 200), order=1)
    assert len(moc.flatten()) < len(moc_complement.flatten())

    moc = _generate_ra_strip_moc(ra_range=(200, 50), order=1)
    moc_complement = _generate_ra_strip_moc(ra_range=(50, 200), order=1)
    assert len(moc_complement.flatten()) < len(moc.flatten())


def test_box_filter_dec(small_sky_order1_catalog):
    # The catalog pixels are distributed around the [-90,0] degree range.
    filtered_catalog = small_sky_order1_catalog.filter_by_box(dec=(10, 20))
    assert len(filtered_catalog.get_healpix_pixels()) == 0
    assert len(filtered_catalog.pixel_tree) == 0
    assert filtered_catalog.catalog_info.total_rows == 0

    filtered_catalog_1 = small_sky_order1_catalog.filter_by_box(dec=(-10, 10))
    filtered_pixels_1 = filtered_catalog_1.get_healpix_pixels()
    assert filtered_pixels_1 == [HealpixPixel(1, 47)]
    assert (1, 47) in filtered_catalog_1.pixel_tree
    assert len(filtered_catalog_1.pixel_tree) == 1
    assert filtered_catalog_1.catalog_info.total_rows == 0

    filtered_catalog_2 = small_sky_order1_catalog.filter_by_box(dec=(-30, -20))
    filtered_pixels_2 = filtered_catalog_2.get_healpix_pixels()
    assert filtered_pixels_2 == [HealpixPixel(1, 45), HealpixPixel(1, 46), HealpixPixel(1, 47)]
    assert (1, 45) in filtered_catalog_2.pixel_tree
    assert (1, 46) in filtered_catalog_2.pixel_tree
    assert (1, 47) in filtered_catalog_2.pixel_tree
    assert len(filtered_catalog_2.pixel_tree) == 3
    assert filtered_catalog_2.catalog_info.total_rows == 0


def test_box_filter_ra_and_dec(small_sky_order1_catalog):
    # The catalog pixels are distributed around the [-90,0] degree range.
    filtered_catalog = small_sky_order1_catalog.filter_by_box(ra=(280, 300), dec=(-30, -20))
    filtered_pixels = filtered_catalog.get_healpix_pixels()

    assert len(filtered_pixels) == 2
    assert filtered_pixels == [HealpixPixel(1, 46), HealpixPixel(1, 47)]

    assert (1, 46) in filtered_catalog.pixel_tree
    assert (1, 47) in filtered_catalog.pixel_tree
    assert len(filtered_catalog.pixel_tree.pixels[1]) == 2
    assert filtered_catalog.catalog_info.total_rows == 0

    # Check that the previous filter is the same as intersecting the ra and dec filters
    filtered_catalog_ra = small_sky_order1_catalog.filter_by_box(ra=(280, 300))
    filtered_catalog_dec = small_sky_order1_catalog.filter_by_box(dec=(-30, -20))
    filtered_catalog_ra_pixels = filtered_catalog_ra.get_healpix_pixels()
    filtered_catalog_dec_pixels = filtered_catalog_dec.get_healpix_pixels()
    intersected_pixels = [
        pixel for pixel in filtered_catalog_ra_pixels if pixel in filtered_catalog_dec_pixels
    ]
    assert filtered_pixels == intersected_pixels


def test_box_filter_empty(small_sky_order1_catalog):
    # It is very difficult to get an empty set of HEALPix with ra for this test catalog
    # as its pixels are very close to the South Pole (dec of -90 degrees). In order 1,
    # they are very large in area, and easily overlap with any ra region.
    filtered_catalog = small_sky_order1_catalog.filter_by_box(ra=(0, 10))
    assert len(filtered_catalog.get_healpix_pixels()) == 2
    assert len(filtered_catalog.pixel_tree) == 2

    filtered_catalog = small_sky_order1_catalog.filter_by_box(dec=(10, 20))
    assert len(filtered_catalog.get_healpix_pixels()) == 0
    assert len(filtered_catalog.pixel_tree) == 0

    filtered_catalog = small_sky_order1_catalog.filter_by_box(ra=(40, 50), dec=(10, 20))
    assert len(filtered_catalog.get_healpix_pixels()) == 0
    assert len(filtered_catalog.pixel_tree) == 0


def test_box_filter_invalid_args(small_sky_order1_catalog):
    # Some declination values are out of the [-90,90] bounds
    with pytest.raises(ValueError, match=ValidatorsErrors.INVALID_DEC):
        small_sky_order1_catalog.filter_by_box(ra=(0, 30), dec=(-100, -70))

    # Declination values should be in ascending order
    with pytest.raises(ValueError, match=ValidatorsErrors.INVALID_RADEC_RANGE):
        small_sky_order1_catalog.filter_by_box(dec=(0, -10))

    # There are ranges are defined with more than 2 values
    with pytest.raises(ValueError, match=ValidatorsErrors.INVALID_RADEC_RANGE):
        small_sky_order1_catalog.filter_by_box(ra=(0, 30), dec=(-30, -40, 10))
    with pytest.raises(ValueError, match=ValidatorsErrors.INVALID_RADEC_RANGE):
        small_sky_order1_catalog.filter_by_box(ra=(0, 30, 40), dec=(-40, 10))

    # The range values coincide (for ra, values are wrapped)
    with pytest.raises(ValueError, match=ValidatorsErrors.INVALID_RADEC_RANGE):
        small_sky_order1_catalog.filter_by_box(ra=(100, 100))
    with pytest.raises(ValueError, match=ValidatorsErrors.INVALID_RADEC_RANGE):
        small_sky_order1_catalog.filter_by_box(ra=(0, 360))
    with pytest.raises(ValueError, match=ValidatorsErrors.INVALID_RADEC_RANGE):
        small_sky_order1_catalog.filter_by_box(dec=(50, 50))

    # No range values were provided
    with pytest.raises(ValueError, match=ValidatorsErrors.INVALID_RADEC_RANGE):
        small_sky_order1_catalog.filter_by_box(ra=None, dec=None)


def test_empty_directory(tmp_path, catalog_info_data):
    """Test loading empty or incomplete data"""
    ## Path doesn't exist
    with pytest.raises(FileNotFoundError):
        read_hats(os.path.join("path", "empty"))

    catalog_path = tmp_path / "empty"
    os.makedirs(catalog_path, exist_ok=True)

    ## Path exists but there's nothing there
    with pytest.raises(FileNotFoundError, match="properties file"):
        read_hats(catalog_path)

    ## catalog_info file exists - getting closer
    properties = TableProperties(**catalog_info_data)
    properties.to_properties_file(catalog_path)

    with pytest.raises(FileNotFoundError, match="metadata"):
        read_hats(catalog_path)

    ## Now we create the needed _metadata and everything is right.
    part_info = PartitionInfo.from_healpix([HealpixPixel(0, 11)])
    part_info.write_to_metadata_files(catalog_path=catalog_path)

    with pytest.warns(UserWarning, match="slow"):
        catalog = read_hats(catalog_path)
    assert catalog.catalog_name == "test_name"


def test_generate_negative_tree_pixels(small_sky_order1_catalog):
    """Test generate_negative_tree_pixels on a basic catalog."""
    expected_pixels = [
        HealpixPixel(0, 0),
        HealpixPixel(0, 1),
        HealpixPixel(0, 2),
        HealpixPixel(0, 3),
        HealpixPixel(0, 4),
        HealpixPixel(0, 5),
        HealpixPixel(0, 6),
        HealpixPixel(0, 7),
        HealpixPixel(0, 8),
        HealpixPixel(0, 9),
        HealpixPixel(0, 10),
    ]

    negative_tree = small_sky_order1_catalog.generate_negative_tree_pixels()

    assert negative_tree == expected_pixels


def test_generate_negative_tree_pixels_order_0(small_sky_catalog):
    """Test generate_negative_tree_pixels on a catalog with only order 0 pixels."""
    expected_pixels = [
        HealpixPixel(0, 0),
        HealpixPixel(0, 1),
        HealpixPixel(0, 2),
        HealpixPixel(0, 3),
        HealpixPixel(0, 4),
        HealpixPixel(0, 5),
        HealpixPixel(0, 6),
        HealpixPixel(0, 7),
        HealpixPixel(0, 8),
        HealpixPixel(0, 9),
        HealpixPixel(0, 10),
    ]

    negative_tree = small_sky_catalog.generate_negative_tree_pixels()

    assert negative_tree == expected_pixels


def test_generate_negative_tree_pixels_multi_order(small_sky_order1_catalog):
    """Test generate_negative_tree_pixels on a catalog with
    missing pixels on multiple order.
    """
    # remove one of the order 1 pixels from the catalog.
    nodes = small_sky_order1_catalog.pixel_tree.get_healpix_pixels()
    small_sky_order1_catalog.pixel_tree = PixelTree.from_healpix(nodes[1:])

    expected_pixels = [
        HealpixPixel(0, 0),
        HealpixPixel(0, 1),
        HealpixPixel(0, 2),
        HealpixPixel(0, 3),
        HealpixPixel(0, 4),
        HealpixPixel(0, 5),
        HealpixPixel(0, 6),
        HealpixPixel(0, 7),
        HealpixPixel(0, 8),
        HealpixPixel(0, 9),
        HealpixPixel(0, 10),
        HealpixPixel(1, 44),
    ]

    negative_tree = small_sky_order1_catalog.generate_negative_tree_pixels()

    assert negative_tree == expected_pixels



"""Tests of file IO (reads and writes)"""

import shutil

import pandas as pd
import pyarrow as pa
import pytest

from hats.io import file_io, paths
from hats.io.parquet_metadata import (
    get_healpix_pixel_from_metadata,
    read_row_group_fragments,
    row_group_stat_single_value,
    write_parquet_metadata,
)
from hats.pixel_math.healpix_pixel import HealpixPixel


def test_write_parquet_metadata(tmp_path, small_sky_dir, small_sky_schema, check_parquet_schema):
    """Copy existing catalog and create new metadata files for it"""
    catalog_base_dir = tmp_path / "catalog"
    shutil.copytree(
        small_sky_dir,
        catalog_base_dir,
    )
    total_rows = write_parquet_metadata(catalog_base_dir)
    assert total_rows == 131
    check_parquet_schema(catalog_base_dir / "dataset" / "_metadata", small_sky_schema)
    ## _common_metadata has 0 row groups
    check_parquet_schema(
        catalog_base_dir / "dataset" / "_common_metadata",
        small_sky_schema,
        0,
    )
    ## Re-write - should still have the same properties.
    total_rows = write_parquet_metadata(catalog_base_dir)
    assert total_rows == 131
    check_parquet_schema(catalog_base_dir / "dataset" / "_metadata", small_sky_schema)
    ## _common_metadata has 0 row groups
    check_parquet_schema(
        catalog_base_dir / "dataset" / "_common_metadata",
        small_sky_schema,
        0,
    )


def test_write_parquet_metadata_order1(
    tmp_path, small_sky_order1_dir, small_sky_schema, check_parquet_schema
):
    """Copy existing catalog and create new metadata files for it,
    using a catalog with multiple files."""
    temp_path = tmp_path / "catalog"
    shutil.copytree(
        small_sky_order1_dir,
        temp_path,
    )

    total_rows = write_parquet_metadata(temp_path)
    assert total_rows == 131
    ## 4 row groups for 4 partitioned parquet files
    check_parquet_schema(
        temp_path / "dataset" / "_metadata",
        small_sky_schema,
        4,
    )
    ## _common_metadata has 0 row groups
    check_parquet_schema(
        temp_path / "dataset" / "_common_metadata",
        small_sky_schema,
        0,
    )


def test_write_parquet_metadata_sorted(
    tmp_path, small_sky_order1_dir, small_sky_schema, check_parquet_schema
):
    """Copy existing catalog and create new metadata files for it,
    using a catalog with multiple files."""
    temp_path = tmp_path / "catalog"
    shutil.copytree(
        small_sky_order1_dir,
        temp_path,
    )

    total_rows = write_parquet_metadata(temp_path)
    assert total_rows == 131
    ## 4 row groups for 4 partitioned parquet files
    check_parquet_schema(
        temp_path / "dataset" / "_metadata",
        small_sky_schema,
        4,
    )
    ## _common_metadata has 0 row groups
    check_parquet_schema(
        temp_path / "dataset" / "_common_metadata",
        small_sky_schema,
        0,
    )


def test_write_index_parquet_metadata(tmp_path, check_parquet_schema):
    """Create an index-like catalog, and test metadata creation."""
    temp_path = tmp_path / "index"

    index_parquet_path = temp_path / "dataset" / "Parts=0" / "part_000_of_001.parquet"
    file_io.make_directory(temp_path / "dataset" / "Parts=0")
    basic_index = pd.DataFrame({"_healpix_29": [4000, 4001], "ps1_objid": [700, 800]})
    file_io.write_dataframe_to_parquet(basic_index, index_parquet_path)

    index_catalog_parquet_metadata = pa.schema(
        [
            pa.field("_healpix_29", pa.int64()),
            pa.field("ps1_objid", pa.int64()),
        ]
    )

    total_rows = write_parquet_metadata(temp_path, order_by_healpix=False)
    assert total_rows == 2

    check_parquet_schema(tmp_path / "index" / "dataset" / "_metadata", index_catalog_parquet_metadata)
    ## _common_metadata has 0 row groups
    check_parquet_schema(
        tmp_path / "index" / "dataset" / "_common_metadata",
        index_catalog_parquet_metadata,
        0,
    )


def test_row_group_fragments(small_sky_order1_dir):
    partition_info_file = paths.get_parquet_metadata_pointer(small_sky_order1_dir)

    num_row_groups = 0
    for _ in read_row_group_fragments(partition_info_file):
        num_row_groups += 1

    assert num_row_groups == 4


def test_row_group_fragments_with_dir(small_sky_order1_dir):
    num_row_groups = 0
    for _ in read_row_group_fragments(small_sky_order1_dir):
        num_row_groups += 1

    assert num_row_groups == 4


def test_row_group_stats(small_sky_dir):
    partition_info_file = paths.get_parquet_metadata_pointer(small_sky_dir)
    first_row_group = next(read_row_group_fragments(partition_info_file))

    assert row_group_stat_single_value(first_row_group, "Norder") == 0
    assert row_group_stat_single_value(first_row_group, "Npix") == 11

    with pytest.raises(ValueError, match="doesn't have expected key"):
        row_group_stat_single_value(first_row_group, "NOT HERE")

    with pytest.raises(ValueError, match="stat min != max"):
        row_group_stat_single_value(first_row_group, "ra")


def test_get_healpix_pixel_from_metadata(small_sky_dir):
    partition_info_file = paths.get_parquet_metadata_pointer(small_sky_dir)
    single_metadata = file_io.read_parquet_metadata(partition_info_file)
    pixel = get_healpix_pixel_from_metadata(single_metadata)
    assert pixel == HealpixPixel(0, 11)


def test_get_healpix_pixel_from_metadata_min_max(tmp_path):
    good_healpix_dataframe = pd.DataFrame({"data": [0, 1], "Norder": [1, 1], "Npix": [44, 44]})
    metadata_filename = tmp_path / "non_healpix_metadata.parquet"
    good_healpix_dataframe.to_parquet(metadata_filename)
    single_metadata = file_io.read_parquet_metadata(metadata_filename)
    pixel = get_healpix_pixel_from_metadata(single_metadata)
    assert pixel == HealpixPixel(1, 44)

    non_healpix_dataframe = pd.DataFrame({"data": [0, 1], "Npix": [45, 44]})
    non_healpix_dataframe.to_parquet(metadata_filename)
    single_metadata = file_io.read_parquet_metadata(metadata_filename)
    with pytest.raises(ValueError, match="Npix stat min != max"):
        get_healpix_pixel_from_metadata(single_metadata)

    non_healpix_dataframe = pd.DataFrame({"data": [0, 1], "Norder": [5, 6]})
    non_healpix_dataframe.to_parquet(metadata_filename)
    single_metadata = file_io.read_parquet_metadata(metadata_filename)
    with pytest.raises(ValueError, match="Norder stat min != max"):
        get_healpix_pixel_from_metadata(single_metadata)


def test_get_healpix_pixel_from_metadata_fail(tmp_path):
    empty_dataframe = pd.DataFrame()
    metadata_filename = tmp_path / "empty_metadata.parquet"
    empty_dataframe.to_parquet(metadata_filename)
    single_metadata = file_io.read_parquet_metadata(metadata_filename)
    with pytest.raises(ValueError, match="empty table"):
        get_healpix_pixel_from_metadata(single_metadata)

    non_healpix_dataframe = pd.DataFrame({"data": [0], "Npix": [45]})
    metadata_filename = tmp_path / "non_healpix_metadata.parquet"
    non_healpix_dataframe.to_parquet(metadata_filename)
    single_metadata = file_io.read_parquet_metadata(metadata_filename)
    with pytest.raises(ValueError, match="missing Norder"):
        get_healpix_pixel_from_metadata(single_metadata)


def test_get_healpix_pixel_from_metadata_columns(tmp_path):
    """Test fetching the healpix pixel from columns with non-default names."""
    non_healpix_dataframe = pd.DataFrame({"data": [1], "Npix": [45], "join_Norder": [2], "join_Npix": [3]})
    metadata_filename = tmp_path / "non_healpix_metadata.parquet"
    non_healpix_dataframe.to_parquet(metadata_filename)
    single_metadata = file_io.read_parquet_metadata(metadata_filename)
    with pytest.raises(ValueError, match="missing Norder"):
        get_healpix_pixel_from_metadata(single_metadata)

    pixel = get_healpix_pixel_from_metadata(single_metadata, norder_column="data")
    assert pixel == HealpixPixel(1, 45)

    pixel = get_healpix_pixel_from_metadata(
        single_metadata, norder_column="join_Norder", npix_column="join_Npix"
    )
    assert pixel == HealpixPixel(2, 3)

    ## People can do silly things!
    pixel = get_healpix_pixel_from_metadata(single_metadata, norder_column="data", npix_column="join_Npix")
    assert pixel == HealpixPixel(1, 3)



import numpy as np

from hats.catalog import Catalog
from hats.pixel_math import HealpixPixel
from hats.pixel_tree.pixel_alignment import PixelAlignment, align_trees, align_with_mocs
from hats.pixel_tree.pixel_tree import PixelTree


def assert_trees_equal(tree1: PixelTree, tree2: PixelTree):
    np.testing.assert_array_equal(tree1.tree, tree2.tree)
    assert tree1.tree_order == tree2.tree_order


def assert_mapping_matches_tree(alignment: PixelAlignment):
    for (_, row), pixel in zip(alignment.pixel_mapping.iterrows(), alignment.pixel_tree.get_healpix_pixels()):
        left_pixel = (
            HealpixPixel(
                row[PixelAlignment.PRIMARY_ORDER_COLUMN_NAME], row[PixelAlignment.PRIMARY_PIXEL_COLUMN_NAME]
            )
            if row[PixelAlignment.PRIMARY_ORDER_COLUMN_NAME] is not None
            else None
        )
        right_pixel = (
            HealpixPixel(
                row[PixelAlignment.JOIN_ORDER_COLUMN_NAME], row[PixelAlignment.JOIN_PIXEL_COLUMN_NAME]
            )
            if row[PixelAlignment.JOIN_ORDER_COLUMN_NAME] is not None
            else None
        )
        aligned_pixel = (
            HealpixPixel(
                row[PixelAlignment.ALIGNED_ORDER_COLUMN_NAME], row[PixelAlignment.ALIGNED_PIXEL_COLUMN_NAME]
            )
            if row[PixelAlignment.ALIGNED_ORDER_COLUMN_NAME] is not None
            else None
        )
        assert pixel == aligned_pixel
        if left_pixel is not None:
            assert aligned_pixel.convert_to_lower_order(aligned_pixel.order - left_pixel.order) == left_pixel
        if right_pixel is not None:
            assert (
                aligned_pixel.convert_to_lower_order(aligned_pixel.order - right_pixel.order) == right_pixel
            )


def test_pixel_tree_alignment_same_tree(pixel_tree_1):
    alignment = align_trees(pixel_tree_1, pixel_tree_1, "inner")
    assert_trees_equal(pixel_tree_1, alignment.pixel_tree)
    assert_mapping_matches_tree(alignment)


def test_pixel_tree_alignment_inner(pixel_tree_2, pixel_tree_3, aligned_trees_2_3_inner):
    alignment = align_trees(pixel_tree_2, pixel_tree_3, "inner")
    assert_trees_equal(alignment.pixel_tree, aligned_trees_2_3_inner)
    assert_mapping_matches_tree(alignment)


def test_pixel_tree_alignment_left(pixel_tree_2, pixel_tree_3, aligned_trees_2_3_left):
    alignment = align_trees(pixel_tree_2, pixel_tree_3, "left")
    assert_trees_equal(alignment.pixel_tree, aligned_trees_2_3_left)
    assert_mapping_matches_tree(alignment)


def test_pixel_tree_alignment_right(pixel_tree_2, pixel_tree_3, aligned_trees_2_3_right):
    alignment = align_trees(pixel_tree_2, pixel_tree_3, "right")
    assert_trees_equal(alignment.pixel_tree, aligned_trees_2_3_right)
    assert_mapping_matches_tree(alignment)


def test_pixel_tree_alignment_outer(pixel_tree_2, pixel_tree_3, aligned_trees_2_3_outer):
    alignment = align_trees(pixel_tree_2, pixel_tree_3, "outer")
    assert_trees_equal(alignment.pixel_tree, aligned_trees_2_3_outer)
    assert_mapping_matches_tree(alignment)


def test_pixel_tree_alignment_with_moc_inner(pixel_tree_2, pixel_tree_3, aligned_trees_2_3_inner):
    moc_pixels = aligned_trees_2_3_inner.get_healpix_pixels()[:-3]
    correct_aligned_tree = PixelTree.from_healpix(moc_pixels)
    moc = correct_aligned_tree.to_moc()
    alignment = align_with_mocs(pixel_tree_2, pixel_tree_3, moc, pixel_tree_3.to_moc())
    assert_trees_equal(alignment.pixel_tree, correct_aligned_tree)
    assert_mapping_matches_tree(alignment)
    alignment = align_with_mocs(pixel_tree_2, pixel_tree_3, pixel_tree_2.to_moc(), moc)
    assert_trees_equal(alignment.pixel_tree, correct_aligned_tree)
    assert_mapping_matches_tree(alignment)


def test_pixel_tree_alignment_with_moc_same(pixel_tree_2, pixel_tree_3, aligned_trees_2_3_inner):
    alignment = align_with_mocs(pixel_tree_2, pixel_tree_3, pixel_tree_2.to_moc(), pixel_tree_3.to_moc())
    assert_trees_equal(alignment.pixel_tree, aligned_trees_2_3_inner)
    assert_mapping_matches_tree(alignment)


def test_pixel_tree_alignment_with_moc_left(pixel_tree_2, pixel_tree_3, aligned_trees_2_3_left):
    moc_pixels = aligned_trees_2_3_left.get_healpix_pixels()[:-3]
    correct_aligned_tree = PixelTree.from_healpix(moc_pixels)
    moc = correct_aligned_tree.to_moc()

    alignment = align_with_mocs(pixel_tree_2, pixel_tree_3, moc, pixel_tree_3.to_moc(), alignment_type="left")
    assert_trees_equal(alignment.pixel_tree, correct_aligned_tree)
    assert_mapping_matches_tree(alignment)

    alignment = align_with_mocs(pixel_tree_2, pixel_tree_3, pixel_tree_2.to_moc(), moc, alignment_type="left")
    assert_trees_equal(alignment.pixel_tree, aligned_trees_2_3_left)
    assert_mapping_matches_tree(alignment)


def test_pixel_tree_alignment_with_moc_right(pixel_tree_2, pixel_tree_3, aligned_trees_2_3_right):
    moc_pixels = aligned_trees_2_3_right.get_healpix_pixels()[:-3]
    correct_aligned_tree = PixelTree.from_healpix(moc_pixels)
    moc = correct_aligned_tree.to_moc()

    alignment = align_with_mocs(
        pixel_tree_2, pixel_tree_3, moc, pixel_tree_3.to_moc(), alignment_type="right"
    )
    assert_trees_equal(alignment.pixel_tree, aligned_trees_2_3_right)
    assert_mapping_matches_tree(alignment)

    alignment = align_with_mocs(
        pixel_tree_2, pixel_tree_3, pixel_tree_2.to_moc(), moc, alignment_type="right"
    )
    assert_trees_equal(alignment.pixel_tree, correct_aligned_tree)
    assert_mapping_matches_tree(alignment)


def test_pixel_tree_alignment_with_moc_outer(pixel_tree_2, pixel_tree_3, aligned_trees_2_3_outer):
    moc_pixels = aligned_trees_2_3_outer.get_healpix_pixels()[:-3]
    moc = PixelTree.from_healpix(moc_pixels).to_moc()
    # moc doesn't include pixels (1,45), (1,46), (1,47). Outer align with right tree moc includes 46 and 47
    # from right tree
    left_moc_correct_tree = PixelTree.from_healpix(moc_pixels + [HealpixPixel(1, 46), HealpixPixel(1, 47)])
    alignment = align_with_mocs(
        pixel_tree_2, pixel_tree_3, moc, pixel_tree_3.to_moc(), alignment_type="outer"
    )
    assert_trees_equal(alignment.pixel_tree, left_moc_correct_tree)
    assert_mapping_matches_tree(alignment)

    # moc doesn't include pixels (1,45), (1,46), (1,47). Outer align with left tree moc includes 45 and 46
    # from left tree
    right_moc_correct_tree = PixelTree.from_healpix(moc_pixels + [HealpixPixel(1, 45), HealpixPixel(1, 46)])
    alignment = align_with_mocs(
        pixel_tree_2, pixel_tree_3, pixel_tree_2.to_moc(), moc, alignment_type="outer"
    )
    assert_trees_equal(alignment.pixel_tree, right_moc_correct_tree)
    assert_mapping_matches_tree(alignment)

    both_moc_correct_tree = PixelTree.from_healpix(moc_pixels)
    alignment = align_with_mocs(pixel_tree_2, pixel_tree_3, moc, moc, alignment_type="outer")
    assert_trees_equal(alignment.pixel_tree, both_moc_correct_tree)
    assert_mapping_matches_tree(alignment)


def test_catalog_align_inner(pixel_tree_2, pixel_tree_3, aligned_trees_2_3_inner, catalog_info):
    left_cat = Catalog(catalog_info, pixel_tree_2)
    right_cat = Catalog(catalog_info, pixel_tree_3)
    alignment = left_cat.align(right_cat)
    assert_trees_equal(alignment.pixel_tree, aligned_trees_2_3_inner)
    assert_mapping_matches_tree(alignment)

    moc_pixels = aligned_trees_2_3_inner.get_healpix_pixels()[:-3]
    correct_aligned_tree = PixelTree.from_healpix(moc_pixels)
    moc = correct_aligned_tree.to_moc()

    left_cat_with_moc = Catalog(catalog_info, pixel_tree_2, moc=moc)
    alignment = left_cat_with_moc.align(right_cat)
    assert_trees_equal(alignment.pixel_tree, correct_aligned_tree)
    assert_mapping_matches_tree(alignment)

    right_cat_with_moc = Catalog(catalog_info, pixel_tree_3, moc=moc)
    alignment = left_cat.align(right_cat_with_moc)
    assert_trees_equal(alignment.pixel_tree, correct_aligned_tree)
    assert_mapping_matches_tree(alignment)


def test_catalog_align_left(pixel_tree_2, pixel_tree_3, aligned_trees_2_3_left, catalog_info):
    left_cat = Catalog(catalog_info, pixel_tree_2)
    right_cat = Catalog(catalog_info, pixel_tree_3)
    alignment = left_cat.align(right_cat, alignment_type="left")
    assert_trees_equal(alignment.pixel_tree, aligned_trees_2_3_left)
    assert_mapping_matches_tree(alignment)

    moc_pixels = aligned_trees_2_3_left.get_healpix_pixels()[:-3]
    correct_aligned_tree = PixelTree.from_healpix(moc_pixels)
    moc = correct_aligned_tree.to_moc()

    left_cat_with_moc = Catalog(catalog_info, pixel_tree_2, moc=moc)
    alignment = left_cat_with_moc.align(right_cat, alignment_type="left")
    assert_trees_equal(alignment.pixel_tree, correct_aligned_tree)
    assert_mapping_matches_tree(alignment)

    right_cat_with_moc = Catalog(catalog_info, pixel_tree_3, moc=moc)
    alignment = left_cat.align(right_cat_with_moc, alignment_type="left")
    assert_trees_equal(alignment.pixel_tree, aligned_trees_2_3_left)
    assert_mapping_matches_tree(alignment)


def test_catalog_align_right(pixel_tree_2, pixel_tree_3, aligned_trees_2_3_right, catalog_info):
    left_cat = Catalog(catalog_info, pixel_tree_2)
    right_cat = Catalog(catalog_info, pixel_tree_3)
    alignment = left_cat.align(right_cat, alignment_type="right")
    assert_trees_equal(alignment.pixel_tree, aligned_trees_2_3_right)
    assert_mapping_matches_tree(alignment)

    moc_pixels = aligned_trees_2_3_right.get_healpix_pixels()[:-3]
    correct_aligned_tree = PixelTree.from_healpix(moc_pixels)
    moc = correct_aligned_tree.to_moc()

    left_cat_with_moc = Catalog(catalog_info, pixel_tree_2, moc=moc)
    alignment = left_cat_with_moc.align(right_cat, alignment_type="right")
    assert_trees_equal(alignment.pixel_tree, aligned_trees_2_3_right)
    assert_mapping_matches_tree(alignment)

    right_cat_with_moc = Catalog(catalog_info, pixel_tree_3, moc=moc)
    alignment = left_cat.align(right_cat_with_moc, alignment_type="right")
    assert_trees_equal(alignment.pixel_tree, correct_aligned_tree)
    assert_mapping_matches_tree(alignment)


def test_catalog_align_outer(pixel_tree_2, pixel_tree_3, aligned_trees_2_3_outer, catalog_info):
    left_cat = Catalog(catalog_info, pixel_tree_2)
    right_cat = Catalog(catalog_info, pixel_tree_3)
    alignment = left_cat.align(right_cat, alignment_type="outer")
    assert_trees_equal(alignment.pixel_tree, aligned_trees_2_3_outer)
    assert_mapping_matches_tree(alignment)

    moc_pixels = aligned_trees_2_3_outer.get_healpix_pixels()[:-3]
    moc = PixelTree.from_healpix(moc_pixels).to_moc()

    left_moc_correct_tree = PixelTree.from_healpix(moc_pixels + [HealpixPixel(1, 46), HealpixPixel(1, 47)])
    left_cat_with_moc = Catalog(catalog_info, pixel_tree_2, moc=moc)
    alignment = left_cat_with_moc.align(right_cat, alignment_type="outer")
    assert_trees_equal(alignment.pixel_tree, left_moc_correct_tree)
    assert_mapping_matches_tree(alignment)

    right_moc_correct_tree = PixelTree.from_healpix(moc_pixels + [HealpixPixel(1, 45), HealpixPixel(1, 46)])
    right_cat_with_moc = Catalog(catalog_info, pixel_tree_3, moc=moc)
    alignment = left_cat.align(right_cat_with_moc, alignment_type="outer")
    assert_trees_equal(alignment.pixel_tree, right_moc_correct_tree)
    assert_mapping_matches_tree(alignment)

    both_moc_correct_tree = PixelTree.from_healpix(moc_pixels)
    alignment = left_cat_with_moc.align(right_cat_with_moc, alignment_type="outer")
    assert_trees_equal(alignment.pixel_tree, both_moc_correct_tree)
    assert_mapping_matches_tree(alignment)


def test_outer_align_start_0():
    left_tree = PixelTree.from_healpix([HealpixPixel(0, 0)])
    right_tree = PixelTree.from_healpix([HealpixPixel(1, 1)])
    alignment = align_trees(left_tree, right_tree, alignment_type="outer")
    expected_tree = PixelTree.from_healpix(
        [HealpixPixel(1, 0), HealpixPixel(1, 1), HealpixPixel(1, 2), HealpixPixel(1, 3)]
    )
    assert_trees_equal(alignment.pixel_tree, expected_tree)
    assert_mapping_matches_tree(alignment)



"""Tests of partition info functionality"""

import numpy.testing as npt
import pandas as pd
import pytest

from hats.catalog import PartitionInfo
from hats.io import paths
from hats.pixel_math import HealpixPixel


def test_load_partition_info_small_sky(small_sky_dir):
    """Instantiate the partition info for catalog with 1 pixel"""
    partition_info_file = paths.get_parquet_metadata_pointer(small_sky_dir)
    partitions = PartitionInfo.read_from_file(partition_info_file)

    order_pixel_pairs = partitions.get_healpix_pixels()
    assert len(order_pixel_pairs) == 1
    expected = [HealpixPixel(0, 11)]
    assert order_pixel_pairs == expected


def test_load_partition_info_from_metadata(small_sky_dir, small_sky_source_dir, small_sky_source_pixels):
    """Instantiate the partition info for catalogs via the `_metadata` file"""
    metadata_file = paths.get_parquet_metadata_pointer(small_sky_dir)
    partitions = PartitionInfo.read_from_file(metadata_file)
    assert partitions.get_healpix_pixels() == [HealpixPixel(0, 11)]

    metadata_file = paths.get_parquet_metadata_pointer(small_sky_source_dir)
    partitions = PartitionInfo.read_from_file(metadata_file)
    assert partitions.get_healpix_pixels() == small_sky_source_pixels

    partitions = PartitionInfo.read_from_file(metadata_file, strict=True)
    assert partitions.get_healpix_pixels() == small_sky_source_pixels


def test_load_partition_info_from_metadata_fail(tmp_path):
    empty_dataframe = pd.DataFrame()
    metadata_filename = tmp_path / "empty_metadata.parquet"
    empty_dataframe.to_parquet(metadata_filename)
    with pytest.raises(ValueError, match="missing Norder"):
        PartitionInfo.read_from_file(metadata_filename)

    with pytest.raises(ValueError, match="at least one column"):
        PartitionInfo.read_from_file(metadata_filename, strict=True)

    non_healpix_dataframe = pd.DataFrame({"data": [0], "Npix": [45]})
    metadata_filename = tmp_path / "non_healpix_metadata.parquet"
    non_healpix_dataframe.to_parquet(metadata_filename)
    with pytest.raises(ValueError, match="missing Norder"):
        PartitionInfo.read_from_file(metadata_filename)

    with pytest.raises(ValueError, match="empty file path"):
        PartitionInfo.read_from_file(metadata_filename, strict=True)


def test_load_partition_info_from_dir_fail(tmp_path):
    empty_dataframe = pd.DataFrame()
    metadata_filename = tmp_path / "empty_metadata.parquet"
    empty_dataframe.to_parquet(metadata_filename)
    with pytest.raises(FileNotFoundError, match="_metadata or partition info"):
        PartitionInfo.read_from_dir(tmp_path)

    # The file is there, but doesn't have the required content.
    (tmp_path / "dataset").mkdir()
    metadata_filename = tmp_path / "dataset" / "_metadata"
    empty_dataframe.to_parquet(metadata_filename)
    with pytest.warns(UserWarning, match="slow"):
        with pytest.raises(ValueError, match="missing Norder"):
            PartitionInfo.read_from_dir(tmp_path)


def test_load_partition_info_small_sky_order1(small_sky_order1_dir):
    """Instantiate the partition info for catalog with 4 pixels"""
    partition_info_file = paths.get_parquet_metadata_pointer(small_sky_order1_dir)
    partitions = PartitionInfo.read_from_file(partition_info_file)

    order_pixel_pairs = partitions.get_healpix_pixels()
    assert len(order_pixel_pairs) == 4
    expected = [
        HealpixPixel(1, 44),
        HealpixPixel(1, 45),
        HealpixPixel(1, 46),
        HealpixPixel(1, 47),
    ]
    assert order_pixel_pairs == expected


def test_load_partition_no_file(tmp_path):
    wrong_path = tmp_path / "_metadata"
    with pytest.raises(FileNotFoundError):
        PartitionInfo.read_from_file(wrong_path)

    wrong_path = tmp_path / "partition_info.csv"
    with pytest.raises(FileNotFoundError):
        PartitionInfo.read_from_csv(wrong_path)


def test_get_highest_order(small_sky_order1_dir):
    """test the `get_highest_order` method"""
    partition_info_file = paths.get_parquet_metadata_pointer(small_sky_order1_dir)
    partitions = PartitionInfo.read_from_file(partition_info_file)

    highest_order = partitions.get_highest_order()

    assert highest_order == 1


def test_calculate_fractional_coverage(small_sky_order1_dir):
    """test the `calculate_fractional_coverage` method"""
    partition_info_file = paths.get_parquet_metadata_pointer(small_sky_order1_dir)
    partitions = PartitionInfo.read_from_file(partition_info_file)

    fractional_coverage = partitions.calculate_fractional_coverage()

    assert fractional_coverage == pytest.approx(0.083, abs=0.001)


def test_write_to_file(tmp_path, small_sky_pixels):
    """Write out the partition info to file and make sure we can read it again."""
    partition_info_pointer = paths.get_partition_info_pointer(tmp_path)
    partition_info = PartitionInfo.from_healpix(small_sky_pixels)
    partition_info.write_to_file(partition_info_pointer)

    new_partition_info = PartitionInfo.read_from_csv(partition_info_pointer)

    # We're not using parquet metadata, so we don't force a re-sorting.
    assert partition_info.get_healpix_pixels() == new_partition_info.get_healpix_pixels()


def test_write_to_file_sorted(tmp_path, pixel_list_depth_first, pixel_list_breadth_first):
    """Write out the partition info to file and make sure that it's sorted by breadth-first healpix,
    even though the original pixel list is in Norder-major sorting (depth-first)."""
    partition_info = PartitionInfo.from_healpix(pixel_list_depth_first)
    npt.assert_array_equal(pixel_list_depth_first, partition_info.get_healpix_pixels())
    total_rows = partition_info.write_to_metadata_files(tmp_path)
    assert total_rows == 9

    partition_info_pointer = paths.get_parquet_metadata_pointer(tmp_path)
    new_partition_info = PartitionInfo.read_from_file(partition_info_pointer)

    npt.assert_array_equal(pixel_list_breadth_first, new_partition_info.get_healpix_pixels())


def test_load_partition_info_from_dir_and_write(tmp_path, pixel_list_depth_first):
    partition_info = PartitionInfo.from_healpix(pixel_list_depth_first)

    ## Path arguments are required if the info was not created from a `read_from_dir` call
    with pytest.raises(ValueError):
        partition_info.write_to_file()
    with pytest.raises(ValueError):
        partition_info.write_to_metadata_files()

    partition_info.write_to_file(catalog_path=tmp_path)
    info = PartitionInfo.read_from_dir(tmp_path)

    ## Can write out the partition info CSV by providing:
    ##  - no arguments
    ##  - new catalog directory
    ##  - full path to the csv file
    info.write_to_file()
    info.write_to_file(catalog_path=tmp_path)
    info.write_to_file(partition_info_file=tmp_path / "new_csv.csv")

    ## Can write out the _metadata file by providing:
    ##  - no arguments
    ##  - new catalog directory
    total_rows = info.write_to_metadata_files()
    assert total_rows == 9
    total_rows = info.write_to_metadata_files(catalog_path=tmp_path)
    assert total_rows == 9



import os.path
from pathlib import Path
from typing import List

import pandas as pd
import pyarrow as pa
import pytest

from hats.catalog.association_catalog.partition_join_info import PartitionJoinInfo
from hats.catalog.dataset.table_properties import TableProperties
from hats.inspection.almanac import Almanac
from hats.pixel_math import HealpixPixel

DATA_DIR_NAME = "data"
ALMANAC_DIR_NAME = "almanac"
SMALL_SKY_DIR_NAME = "small_sky"
SMALL_SKY_ORDER1_DIR_NAME = "small_sky_order1"
SMALL_SKY_SOURCE_OBJECT_INDEX_DIR_NAME = "small_sky_source_object_index"

TEST_DIR = os.path.dirname(__file__)

# pylint: disable=missing-function-docstring, redefined-outer-name


@pytest.fixture
def test_data_dir():
    return Path(TEST_DIR) / DATA_DIR_NAME


@pytest.fixture
def almanac_dir(test_data_dir):
    return test_data_dir / ALMANAC_DIR_NAME


@pytest.fixture
def small_sky_dir(test_data_dir):
    return test_data_dir / SMALL_SKY_DIR_NAME


@pytest.fixture
def small_sky_order1_dir(test_data_dir):
    return test_data_dir / SMALL_SKY_ORDER1_DIR_NAME


@pytest.fixture
def small_sky_source_object_index_dir(test_data_dir):
    return test_data_dir / SMALL_SKY_SOURCE_OBJECT_INDEX_DIR_NAME


@pytest.fixture
def catalog_info_data() -> dict:
    return {
        "catalog_name": "test_name",
        "catalog_type": "object",
        "total_rows": 10,
        "ra_column": "ra",
        "dec_column": "dec",
    }


@pytest.fixture
def catalog_info(catalog_info_data) -> TableProperties:
    return TableProperties(**catalog_info_data)


@pytest.fixture
def association_catalog_info_data() -> dict:
    return {
        "catalog_name": "test_name",
        "catalog_type": "association",
        "total_rows": 10,
        "primary_catalog": "small_sky",
        "primary_column": "id",
        "primary_column_association": "id_small_sky",
        "join_catalog": "small_sky_order1",
        "join_column": "id",
        "join_column_association": "id_small_sky_order1",
        "contains_leaf_files": False,
    }


@pytest.fixture
def association_catalog_info(association_catalog_info_data) -> TableProperties:
    return TableProperties(**association_catalog_info_data)


@pytest.fixture
def source_catalog_info() -> dict:
    return {
        "catalog_name": "test_source",
        "catalog_type": "source",
        "total_rows": 100,
        "ra_column": "source_ra",
        "dec_column": "source_dec",
    }


@pytest.fixture
def margin_cache_catalog_info_data() -> dict:
    return {
        "catalog_name": "test_margin",
        "catalog_type": "margin",
        "total_rows": 100,
        "ra_column": "ra",
        "dec_column": "dec",
        "primary_catalog": "test_name",
        "margin_threshold": 0.5,
    }


@pytest.fixture
def margin_catalog_info(margin_cache_catalog_info_data) -> TableProperties:
    return TableProperties(**margin_cache_catalog_info_data)


@pytest.fixture
def small_sky_schema() -> pa.Schema:
    return pa.schema(
        [
            pa.field("id", pa.int64()),
            pa.field("ra", pa.float64()),
            pa.field("dec", pa.float64()),
            pa.field("ra_error", pa.int64()),
            pa.field("dec_error", pa.int64()),
            pa.field("Norder", pa.uint8()),
            pa.field("Dir", pa.uint64()),
            pa.field("Npix", pa.uint64()),
            pa.field("_healpix_29", pa.int64()),
        ]
    )


@pytest.fixture
def small_sky_source_schema() -> pa.Schema:
    return pa.schema(
        [
            pa.field("source_id", pa.int64()),
            pa.field("source_ra", pa.float64()),
            pa.field("source_dec", pa.float64()),
            pa.field("mjd", pa.float64()),
            pa.field("mag", pa.float64()),
            pa.field("band", pa.string()),
            pa.field("object_id", pa.int64()),
            pa.field("object_ra", pa.float64()),
            pa.field("object_dec", pa.float64()),
            pa.field("Norder", pa.uint8()),
            pa.field("Dir", pa.uint64()),
            pa.field("Npix", pa.uint64()),
            pa.field("_healpix_29", pa.int64()),
        ]
    )


@pytest.fixture
def association_catalog_schema() -> pa.Schema:
    return pa.schema(
        [
            pa.field("Norder", pa.int64()),
            pa.field("Npix", pa.int64()),
            pa.field("join_Norder", pa.int64()),
            pa.field("join_Npix", pa.int64()),
        ]
    )


@pytest.fixture
def margin_catalog_schema() -> pa.Schema:
    return pa.schema(
        [
            pa.field("id", pa.int64()),
            pa.field("ra", pa.float64()),
            pa.field("dec", pa.float64()),
            pa.field("ra_error", pa.int64()),
            pa.field("dec_error", pa.int64()),
            pa.field("Norder", pa.uint8()),
            pa.field("Dir", pa.uint64()),
            pa.field("Npix", pa.uint64()),
            pa.field("_healpix_29", pa.int64()),
            pa.field("margin_Norder", pa.uint8()),
            pa.field("margin_Dir", pa.uint64()),
            pa.field("margin_Npix", pa.uint64()),
        ]
    )


@pytest.fixture
def dataset_path(test_data_dir) -> str:
    return test_data_dir / "info_only" / "dataset"


@pytest.fixture
def catalog_path(test_data_dir) -> str:
    return test_data_dir / "info_only" / "catalog"


@pytest.fixture
def margin_catalog_pixels() -> List[HealpixPixel]:
    return [
        HealpixPixel(0, 4),
        HealpixPixel(1, 44),
        HealpixPixel(1, 45),
        HealpixPixel(1, 46),
        HealpixPixel(1, 47),
    ]


@pytest.fixture
def margin_catalog_path(test_data_dir) -> str:
    return test_data_dir / "small_sky_order1_margin"


@pytest.fixture
def catalog_pixels() -> List[HealpixPixel]:
    return [HealpixPixel(1, 0), HealpixPixel(1, 1), HealpixPixel(2, 8)]


@pytest.fixture
def association_catalog_path(test_data_dir) -> str:
    return test_data_dir / "small_sky_to_small_sky_order1"


@pytest.fixture
def small_sky_source_dir(test_data_dir) -> str:
    return test_data_dir / "small_sky_source"


@pytest.fixture
def small_sky_source_pixels():
    """Source catalog pixels"""
    return [
        HealpixPixel(0, 4),
        HealpixPixel(2, 176),
        HealpixPixel(2, 177),
        HealpixPixel(2, 178),
        HealpixPixel(2, 179),
        HealpixPixel(2, 180),
        HealpixPixel(2, 181),
        HealpixPixel(2, 182),
        HealpixPixel(2, 183),
        HealpixPixel(2, 184),
        HealpixPixel(2, 185),
        HealpixPixel(2, 186),
        HealpixPixel(2, 187),
        HealpixPixel(1, 47),
    ]


@pytest.fixture
def association_catalog_partition_join_file(association_catalog_path) -> str:
    return association_catalog_path / "partition_join_info.csv"


@pytest.fixture
def association_catalog_join_pixels() -> pd.DataFrame:
    return pd.DataFrame.from_dict(
        {
            PartitionJoinInfo.PRIMARY_ORDER_COLUMN_NAME: [0, 0, 0, 0],
            PartitionJoinInfo.PRIMARY_PIXEL_COLUMN_NAME: [11, 11, 11, 11],
            PartitionJoinInfo.JOIN_ORDER_COLUMN_NAME: [1, 1, 1, 1],
            PartitionJoinInfo.JOIN_PIXEL_COLUMN_NAME: [44, 45, 46, 47],
        }
    )


@pytest.fixture
def default_almanac(almanac_dir, test_data_dir):
    """Set up default environment variables and fetch default almanac data."""
    os.environ["HATS_ALMANAC_DIR"] = str(almanac_dir)
    os.environ["HATS_DEFAULT_DIR"] = str(test_data_dir)

    return Almanac()



from typing import List

from hats.pixel_math import HealpixPixel
from hats.pixel_tree import PixelAlignment, PixelAlignmentType, align_trees
from hats.pixel_tree.pixel_tree import PixelTree


def get_filtered_pixel_list(pixel_tree: PixelTree, search_tree: PixelTree) -> List[HealpixPixel]:
    """Aligns the catalog pixel tree with another pixel tree of interest, and
    extracts the HEALPix pixels that overlap. This method is useful to obtain
    the pixels intersecting a pre-defined search space.

    Args:
        pixel_tree (PixelTree): The catalog pixel tree
        search_tree (PixelTree): The tree of pixels that defines the search space

    Returns:
        A list of HEALPix pixels that overlap in both trees.
    """
    search_alignment = align_trees(pixel_tree, search_tree, alignment_type=PixelAlignmentType.INNER)
    pixels_df = search_alignment.pixel_mapping[
        [PixelAlignment.PRIMARY_ORDER_COLUMN_NAME, PixelAlignment.PRIMARY_PIXEL_COLUMN_NAME]
    ]
    filtered_pixels_df = pixels_df.drop_duplicates()
    pixel_list = [
        HealpixPixel(order, pixel)
        for order, pixel in zip(
            filtered_pixels_df[PixelAlignment.PRIMARY_ORDER_COLUMN_NAME],
            filtered_pixels_df[PixelAlignment.PRIMARY_PIXEL_COLUMN_NAME],
        )
    ]
    return pixel_list



from __future__ import annotations

from pathlib import Path
from typing import Type

from upath import UPath

from hats.catalog import AssociationCatalog, Catalog, CatalogType, Dataset, MarginCatalog
from hats.catalog.dataset.table_properties import TableProperties
from hats.catalog.index.index_catalog import IndexCatalog

CATALOG_TYPE_TO_CLASS = {
    CatalogType.OBJECT: Catalog,
    CatalogType.SOURCE: Catalog,
    CatalogType.ASSOCIATION: AssociationCatalog,
    CatalogType.INDEX: IndexCatalog,
    CatalogType.MARGIN: MarginCatalog,
}


def read_hats(catalog_path: str | Path | UPath, catalog_type: CatalogType | None = None) -> Dataset:
    """Reads a HATS Catalog from a HATS directory

    Args:
        catalog_path (str): path to the root directory of the catalog
        catalog_type (CatalogType): Default `None`. By default, the type of the catalog is loaded
            from the catalog info and the corresponding object type is returned. Python's type hints
            cannot allow a return type specified by a loaded value, so to use the correct return
            type for type checking, the type of the catalog can be specified here. Use by specifying
            the hats class for that catalog.

    Returns:
        The initialized catalog object
    """
    catalog_type_to_use = (
        _read_dataset_class_from_metadata(catalog_path) if catalog_type is None else catalog_type
    )
    loader = _get_loader_from_catalog_type(catalog_type_to_use)
    return loader.read_hats(catalog_path)


def _read_dataset_class_from_metadata(catalog_base_path: str) -> CatalogType:
    return TableProperties.read_from_dir(catalog_base_path).catalog_type


def _get_loader_from_catalog_type(catalog_type: CatalogType) -> Type[Dataset]:
    if catalog_type not in CATALOG_TYPE_TO_CLASS:
        raise NotImplementedError(f"Cannot load catalog of type {catalog_type}")
    return CATALOG_TYPE_TO_CLASS[catalog_type]



import healpy as hp
import pytest

from hats.catalog import Catalog
from hats.inspection import plot_pixel_list, plot_pixels, plot_points
from hats.loaders import read_hats

# pylint: disable=no-member


@pytest.mark.timeout(10)
def test_generate_projections(small_sky_dir, mocker):
    """Basic test that map data can be generated"""

    cat = read_hats(small_sky_dir)

    projection = "moll"
    mocker.patch("healpy.mollview")
    plot_pixels(cat, projection=projection)
    hp.mollview.assert_called_once()
    hp.mollview.reset_mock()
    plot_points(cat, projection=projection)
    hp.mollview.assert_called_once()

    projection = "gnom"
    mocker.patch("healpy.gnomview")
    plot_pixels(cat, projection=projection)
    hp.gnomview.assert_called_once()
    hp.gnomview.reset_mock()
    plot_points(cat, projection=projection)
    hp.gnomview.assert_called_once()

    projection = "cart"
    mocker.patch("healpy.cartview")
    plot_pixels(cat, projection=projection)
    hp.cartview.assert_called_once()
    hp.cartview.reset_mock()
    plot_points(cat, projection=projection)
    hp.cartview.assert_called_once()

    projection = "orth"
    mocker.patch("healpy.orthview")
    plot_pixels(cat, projection=projection)
    hp.orthview.assert_called_once()
    hp.orthview.reset_mock()
    plot_points(cat, projection=projection)
    hp.orthview.assert_called_once()


def test_generate_map_unknown_projection(small_sky_dir):
    """Test for error with unknown projection type"""

    cat = read_hats(small_sky_dir)
    with pytest.raises(NotImplementedError):
        plot_pixels(cat, projection=None)

    with pytest.raises(NotImplementedError):
        plot_pixels(cat, projection="")

    with pytest.raises(NotImplementedError):
        plot_pixels(cat, projection="projection")


def test_plot_pixel_map_order1(small_sky_order1_dir, mocker):
    """Basic test that map data can be generated"""
    mocker.patch("healpy.mollview")

    cat = read_hats(small_sky_order1_dir)
    plot_pixels(cat)
    hp.mollview.assert_called_once()

    hp.mollview.reset_mock()

    plot_points(cat)
    hp.mollview.assert_called_once()


def test_visualize_in_memory_catalogs(catalog_info, catalog_pixels, mocker):
    """Test behavior of visualization methods for non-on-disk catalogs and pixel data."""
    mocker.patch("healpy.mollview")
    catalog = Catalog(catalog_info, catalog_pixels)
    plot_pixels(catalog)
    hp.mollview.assert_called_once()

    hp.mollview.reset_mock()
    plot_pixel_list(catalog_pixels, plot_title="My special catalog")
    hp.mollview.assert_called_once()

    hp.mollview.reset_mock()
    with pytest.raises(ValueError, match="on disk catalog required"):
        plot_points(catalog)
    hp.mollview.assert_not_called()



from __future__ import annotations

import os
import warnings
from typing import List

import pandas as pd

from hats.catalog.catalog_type import CatalogType
from hats.catalog.dataset.dataset import Dataset
from hats.inspection.almanac_info import AlmanacInfo
from hats.io.file_io import file_pointer
from hats.loaders import read_hats


class Almanac:
    """Single instance of an almanac, and available catalogs within namespaces

    Params:
        include_default_dir:
            include directory indicated in HATS_ALMANAC_DIR
            environment variable. see AlmanacInfo.get_default_dir
        dirs:
            additional directories or files to look for almanac files in. we support a
            few types of input, with different behaviors:

            - ``str`` - a single directory (or file)
            - ``list[str]`` - multiple directories (or files)
            - ``dict[str:str]`` / ``dict[str:list[str]]`` - namespace
              dictionary. for each key in the dictionary, we put all almanac
              entries under a namespace. this is useful if you have name
              collisions e.g. between multiple surveys or user-provided
              catalogs.
    """

    def __init__(self, include_default_dir=True, dirs=None):
        """Create new almanac."""
        self.files = {}
        self.entries = {}
        self.dir_to_catalog_name = {}
        self._init_files(include_default_dir=include_default_dir, dirs=dirs)
        self._init_catalog_objects()
        self._init_catalog_links()

    def _init_files(self, include_default_dir=True, dirs=None):
        """Create a list of all the almanac files we want to add to this instance.

        Each almanac file corresponds to a single catalog.

        Args:
            include_default_dir: include directory indicated in HATS_ALMANAC_DIR
                environment variable. see AlmanacInfo.get_default_dir
            dirs: additional directories to look for almanac files in
        """
        if include_default_dir:
            default_dir = AlmanacInfo.get_default_dir()
            if default_dir:
                self._add_files_to_namespace(default_dir)
        if pd.api.types.is_dict_like(dirs):
            for key, value in dirs.items():
                self._add_files_to_namespace(value, key)
        elif pd.api.types.is_list_like(dirs):
            self._add_files_to_namespace(dirs)
        elif dirs is not None:
            self._add_files_to_namespace(dirs)

    def _add_files_to_namespace(self, directory, namespace=""):
        """Get almanac files within a directory or list of directories.

        Args:
            directory: directory to scan
            namespace: if provided, files in this directory will be in their
                own namespace in the almanac
        """
        if not pd.api.types.is_list_like(directory):
            directory = [directory]

        files = []
        for input_path in directory:
            if os.path.isfile(input_path):
                files.append(input_path)
                continue

            path_contents = file_pointer.get_directory_contents(input_path)
            input_paths = [x for x in path_contents if str(x).endswith(".yml")]
            input_paths.sort()
            files.extend(input_paths)

        if namespace in self.files:
            self.files[namespace].extend(files)
        else:
            self.files[namespace] = files

    def _init_catalog_objects(self):
        """Create (unlinked) almanac info objects for all the files found
        in the previous steps."""
        for namespace, files in self.files.items():
            for file in files:
                catalog_info = AlmanacInfo.from_file(file)
                catalog_info.namespace = namespace
                if namespace:
                    full_name = f"{namespace}:{catalog_info.catalog_name}"
                else:
                    full_name = catalog_info.catalog_name
                if full_name in self.entries:
                    warnings.warn(f"Duplicate catalog name ({full_name}). Try using namespaces.")
                else:
                    self.entries[full_name] = catalog_info
                    self.dir_to_catalog_name[catalog_info.catalog_path] = full_name

    def _init_catalog_links(self):
        """Initialize the links between almanac catalogs.

        For each type of link (e.g. primary or join), look for the catalog in
        the almanac, using whatever text we have. If found, add the object
        to the almanac info as a pointer. Additionally, add the reference to
        the linked object, so catalogs know about each other from either side.
        """
        for catalog_entry in self.entries.values():
            if catalog_entry.catalog_type == CatalogType.OBJECT:
                ## Object currently has no links that start in the catalog.
                pass
            elif catalog_entry.catalog_type == CatalogType.SOURCE:
                ## Source catalogs MAY indicate their primary object catalog.
                if catalog_entry.primary:
                    object_catalog = self._get_linked_catalog(catalog_entry.primary, catalog_entry.namespace)
                    if not object_catalog:
                        warnings.warn(
                            f"source catalog {catalog_entry.catalog_name} missing "
                            f"object catalog {catalog_entry.primary}"
                        )
                    else:
                        catalog_entry.primary_link = object_catalog
                        catalog_entry.objects.append(object_catalog)
                        object_catalog.sources.append(catalog_entry)
            elif catalog_entry.catalog_type == CatalogType.ASSOCIATION:
                ## Association table MUST have a primary and join catalog
                primary_catalog = self._get_linked_catalog(catalog_entry.primary, catalog_entry.namespace)
                if not primary_catalog:
                    warnings.warn(
                        f"association table {catalog_entry.catalog_name} missing "
                        f"primary catalog {catalog_entry.primary}"
                    )
                else:
                    catalog_entry.primary_link = primary_catalog
                    primary_catalog.associations.append(catalog_entry)

                join_catalog = self._get_linked_catalog(
                    catalog_entry.join,
                    catalog_entry.namespace,
                )
                if not join_catalog:
                    warnings.warn(
                        f"association table {catalog_entry.catalog_name} missing "
                        f"join catalog {catalog_entry.join}"
                    )
                else:
                    catalog_entry.join_link = join_catalog
                    join_catalog.associations_right.append(catalog_entry)
            elif catalog_entry.catalog_type == CatalogType.MARGIN:
                ## Margin catalogs MUST have a primary catalog
                primary_catalog = self._get_linked_catalog(catalog_entry.primary, catalog_entry.namespace)
                if not primary_catalog:
                    warnings.warn(
                        f"margin table {catalog_entry.catalog_name} missing "
                        f"primary catalog {catalog_entry.primary}"
                    )
                else:
                    catalog_entry.primary_link = primary_catalog
                    primary_catalog.margins.append(catalog_entry)
            elif catalog_entry.catalog_type == CatalogType.INDEX:
                ## Index tables MUST have a primary catalog
                primary_catalog = self._get_linked_catalog(catalog_entry.primary, catalog_entry.namespace)
                if not primary_catalog:
                    warnings.warn(
                        f"index table {catalog_entry.catalog_name} missing "
                        f"primary catalog {catalog_entry.primary}"
                    )
                else:
                    catalog_entry.primary_link = primary_catalog
                    primary_catalog.indexes.append(catalog_entry)
            else:  # pragma: no cover
                warnings.warn(f"Unknown catalog type {catalog_entry.catalog_type}")

    def _get_linked_catalog(self, linked_text, namespace) -> AlmanacInfo | None:
        """Find a catalog to be used for linking catalogs within the almanac.

        e.g. for an association table, we will have a primary and join catalog.
        the association catalog is "receiving" the link of primary catalog info,
        and a link of join catalog info.

        Args:
            linked_text: text provided for the linked catalog. this could take
                a few different forms:

                - empty or None (returns None)
                - short name of a catalog
                - namespaced name of a catalog
                - full path to a catalog base directory
                - path to a catalog base directory, with environment variables
            namespace: the namespace in the catalog **receiving** the link.
                this is used to resolve the linked_text argument, so if you're
                relying on namespaces, the receiving and linking catalog should
                be in the same namespace
        Returns:
            almanac info for the linked catalog, if found
        """
        resolved_path = os.path.expandvars(linked_text)
        if linked_text in self.dir_to_catalog_name:  # pragma: no cover
            linked_text = self.dir_to_catalog_name[linked_text]
        elif resolved_path in self.dir_to_catalog_name:
            linked_text = self.dir_to_catalog_name[resolved_path]

        resolved_name = linked_text
        if resolved_name not in self.entries:
            resolved_name = f"{namespace}:{linked_text}"
            if resolved_name not in self.entries:
                return None
        return self.entries[resolved_name]

    def catalogs(self, include_deprecated=False, types: List[str] | None = None):
        """Get names of catalogs in the almanac, matching the provided conditions.

        Catalogs must meet all criteria provided in order to be returned (e.g.
        the criteria are ANDED together).

        Args:
            include_deprecated: include catalogs which contain some text in their
                ``deprecated`` field.
            types: include ONLY catalogs within the list of provided types.
        """
        selected = []
        for full_name, catalog_info in self.entries.items():
            include = True
            if not include_deprecated and catalog_info.deprecated:
                include = False
            if types and catalog_info.catalog_type not in types:
                include = False

            if include:
                selected.append(full_name)
        return selected

    def get_almanac_info(self, catalog_name: str) -> AlmanacInfo:
        """Fetch the almanac info for a single catalog."""
        return self.entries[catalog_name]

    def get_catalog(self, catalog_name: str) -> Dataset:
        """Fetch the fully-populated hats metadata for the catalog name.

        This will load the ``catalog_info.join`` and other relevant metadata files
        from disk."""
        return read_hats(self.get_almanac_info(catalog_name=catalog_name).catalog_path)



from __future__ import annotations

from collections.abc import Sequence
from typing import List

import numpy as np
from mocpy import MOC

from hats.pixel_math import HealpixInputTypes, HealpixPixel
from hats.pixel_math.healpix_pixel_convertor import get_healpix_tuple
from hats.pixel_math.healpix_pixel_function import get_pixels_from_intervals


class PixelTree:
    """Sparse Quadtree of HEALPix pixels that make up the HATS catalog

    This class stores each node in the tree, with leaf nodes corresponding to pixels with data
    files.

    There are a number of methods in this class which allow for quickly navigating through the
    tree and performing operations to filter the pixels in the catalog.

    Attributes:
        pixels: Nested dictionary of pixel nodes stored in the tree. Indexed by HEALPix
            order then pixel number
    """

    def __init__(self, tree: np.ndarray, order: int) -> None:
        """Initialises a tree object from the nodes in the tree

        Args:
            tree (np.ndarray): sorted array of intervals that represent each pixel in the tree
            order (int): HEALPix order of the pixel numbers in the intervals
        """
        self.tree_order = order
        self.tree = tree

        if not np.all((self.tree.T[0, 1:] - self.tree.T[1, :-1]) >= 0):
            raise ValueError("Invalid Catalog: Tree contains overlapping pixels")

        self.pixels = get_pixels_from_intervals(self.tree, self.tree_order)

    def __len__(self):
        """Gets the number of nodes in the tree

        Returns:
            The number of nodes in the tree
        """
        return len(self.tree)

    def contains(self, pixel: HealpixInputTypes) -> bool:
        """Check if tree contains a node at a given order and pixel

        Args:
            pixel: HEALPix pixel to check. Either of type `HealpixPixel`
                or a tuple of (order, pixel)

        Returns:
            True if the tree contains the pixel, False if not
        """
        (order, pixel) = get_healpix_tuple(pixel)
        if order > self.tree_order:
            return False
        d_order = self.tree_order - order
        pixel_at_tree_order = pixel << 2 * d_order
        index = np.searchsorted(self.tree.T[1], pixel_at_tree_order, side="right")
        if index >= len(self.pixels):
            return False
        is_same_order = self.pixels[index][0] == order
        return pixel_at_tree_order == self.tree[index][0] and is_same_order

    def __contains__(self, item):
        return self.contains(item)

    def get_max_depth(self) -> int:
        """Get the max depth (or highest healpix order) represented in the list of pixels.

        Returns:
            max depth (or highest healpix order) of the pixels in the tree
        """
        return np.max(self.pixels.T[0])

    def get_healpix_pixels(self) -> List[HealpixPixel]:
        """Creates a list of HealpixPixels in the tree

        Returns (List[HealpixPixel]):
            A list of the HEALPix pixels in the tree
        """
        return [HealpixPixel(p[0], p[1]) for p in self.pixels]

    def to_moc(self) -> MOC:
        """Returns the MOC object that covers the same pixels as the tree"""
        return MOC.from_healpix_cells(self.pixels.T[1], self.pixels.T[0], self.tree_order)

    @classmethod
    def from_healpix(cls, healpix_pixels: Sequence[HealpixInputTypes], tree_order=None) -> PixelTree:
        """Build a tree from a list of constituent healpix pixels

        Args:
            healpix_pixels: list of healpix pixels
            tree_order (int): (Default = None) order to generate the tree at. If None, will use the highest
                order from input pixels

        Returns:
            The pixel tree with the leaf pixels specified in the list
        """
        if len(healpix_pixels) == 0:
            return PixelTree(np.empty((0, 2), dtype=np.int64), 0)

        pixel_tuples = [get_healpix_tuple(p) for p in healpix_pixels]
        pixel_array = np.array(pixel_tuples).T
        orders = pixel_array[0]
        pixels = pixel_array[1]
        max_order = np.max(orders) if tree_order is None else tree_order
        starts = pixels * 4 ** (max_order - orders)
        ends = (pixels + 1) * 4 ** (max_order - orders)
        result = np.vstack((starts, ends)).T
        result.sort(axis=0)
        return cls(result, max_order)



import os

import pytest

from hats.catalog.dataset.dataset import Dataset


def test_read_hats(dataset_path):
    dataset = Dataset.read_hats(dataset_path)
    assert dataset.on_disk
    assert str(dataset.catalog_path) == str(dataset_path)
    assert str(dataset.catalog_base_dir) == str(dataset_path)


def test_read_from_missing_folder(tmp_path):
    wrong_path = tmp_path / "wrong"
    with pytest.raises(FileNotFoundError, match="directory"):
        Dataset.read_hats(wrong_path)


def test_read_from_empty_folder(tmp_path):
    dataset_path = tmp_path / "dat"
    os.makedirs(dataset_path)
    with pytest.raises(FileNotFoundError, match="properties file"):
        Dataset.read_hats(dataset_path)



"""Container class to hold primary-to-join partition metadata"""

from __future__ import annotations

import warnings
from pathlib import Path
from typing import Dict, List

import numpy as np
import pandas as pd
import pyarrow as pa
from upath import UPath

from hats.catalog.partition_info import PartitionInfo
from hats.io import file_io, paths
from hats.io.parquet_metadata import (
    read_row_group_fragments,
    row_group_stat_single_value,
    write_parquet_metadata_for_batches,
)
from hats.pixel_math.healpix_pixel import HealpixPixel


class PartitionJoinInfo:
    """Association catalog metadata with which partitions matches occur in the join"""

    PRIMARY_ORDER_COLUMN_NAME = "Norder"
    PRIMARY_PIXEL_COLUMN_NAME = "Npix"
    JOIN_ORDER_COLUMN_NAME = "join_Norder"
    JOIN_PIXEL_COLUMN_NAME = "join_Npix"

    COLUMN_NAMES = [
        PRIMARY_ORDER_COLUMN_NAME,
        PRIMARY_PIXEL_COLUMN_NAME,
        JOIN_ORDER_COLUMN_NAME,
        JOIN_PIXEL_COLUMN_NAME,
    ]

    def __init__(self, join_info_df: pd.DataFrame, catalog_base_dir: str = None) -> None:
        self.data_frame = join_info_df
        self.catalog_base_dir = catalog_base_dir
        self._check_column_names()

    def _check_column_names(self):
        for column in self.COLUMN_NAMES:
            if column not in self.data_frame.columns:
                raise ValueError(f"join_info_df does not contain column {column}")

    def primary_to_join_map(self) -> Dict[HealpixPixel, List[HealpixPixel]]:
        """Generate a map from a single primary pixel to one or more pixels in the join catalog.

        Lots of cute comprehension is happening here, so watch out!
        We create tuple of (primary order/pixel) and [array of tuples of (join order/pixel)]

        Returns:
            dictionary mapping (primary order/pixel) to [array of (join order/pixel)]
        """
        primary_map = self.data_frame.groupby(
            [self.PRIMARY_ORDER_COLUMN_NAME, self.PRIMARY_PIXEL_COLUMN_NAME], group_keys=True
        )
        primary_to_join = [
            (
                HealpixPixel(int(primary_pixel[0]), int(primary_pixel[1])),
                [
                    HealpixPixel(int(object_elem[0]), int(object_elem[1]))
                    for object_elem in join_group.dropna().to_numpy().T[2:4].T
                ],
            )
            for primary_pixel, join_group in primary_map
        ]
        ## Treat the array of tuples as a dictionary.
        primary_to_join = dict(primary_to_join)
        return primary_to_join

    def write_to_metadata_files(self, catalog_path: str | Path | UPath | None = None):
        """Generate parquet metadata, using the known joint partitions.

        Args:
            catalog_path (UPath): base path for the catalog

        Returns:
            sum of the number of rows in the dataset.

        Raises:
            ValueError: if no path is provided, and could not be inferred.
        """
        if catalog_path is None:
            if self.catalog_base_dir is None:
                raise ValueError("catalog_path is required if join info was not loaded from a directory")
            catalog_path = self.catalog_base_dir

        batches = [
            [
                pa.RecordBatch.from_arrays(
                    [
                        [primary_pixel.order],
                        [primary_pixel.pixel],
                        [join_pixel.order],
                        [join_pixel.pixel],
                    ],
                    names=self.COLUMN_NAMES,
                )
                for join_pixel in join_pixels
            ]
            for primary_pixel, join_pixels in self.primary_to_join_map().items()
        ]

        return write_parquet_metadata_for_batches(batches, catalog_path)

    def write_to_csv(self, catalog_path: str | Path | UPath | None = None):
        """Write all partition data to CSV files.

        Two files will be written:

            - partition_info.csv - covers all primary catalog pixels, and should match the file structure
            - partition_join_info.csv - covers all pairwise relationships between primary and
              join catalogs.

        Args:
            catalog_path: path to the directory where the
                `partition_join_info.csv` file will be written

        Raises:
            ValueError: if no path is provided, and could not be inferred.
        """
        if catalog_path is None:
            if self.catalog_base_dir is None:
                raise ValueError("catalog_path is required if info was not loaded from a directory")
            catalog_path = self.catalog_base_dir

        partition_join_info_file = paths.get_partition_join_info_pointer(catalog_path)
        file_io.write_dataframe_to_csv(self.data_frame, partition_join_info_file, index=False)

        primary_pixels = self.primary_to_join_map().keys()
        partition_info_pointer = paths.get_partition_info_pointer(catalog_path)
        partition_info = PartitionInfo.from_healpix(primary_pixels)
        partition_info.write_to_file(partition_info_file=partition_info_pointer)

    @classmethod
    def read_from_dir(cls, catalog_base_dir: str | Path | UPath | None = None) -> PartitionJoinInfo:
        """Read partition join info from a file within a hats directory.

        This will look for a `partition_join_info.csv` file, and if not found, will look for
        a `_metadata` file. The second approach is typically slower for large catalogs
        therefore a warning is issued to the user. In internal testing with large catalogs,
        the first approach takes less than a second, while the second can take 10-20 seconds.

        Args:
            catalog_base_dir: path to the root directory of the catalog

        Returns:
            A `PartitionJoinInfo` object with the data from the file

        Raises:
            FileNotFoundError: if neither desired file is found in the catalog_base_dir
        """
        metadata_file = paths.get_parquet_metadata_pointer(catalog_base_dir)
        partition_join_info_file = paths.get_partition_join_info_pointer(catalog_base_dir)
        if file_io.does_file_or_directory_exist(partition_join_info_file):
            pixel_frame = PartitionJoinInfo._read_from_csv(partition_join_info_file)
        elif file_io.does_file_or_directory_exist(metadata_file):
            warnings.warn("Reading partitions from parquet metadata. This is typically slow.")
            pixel_frame = PartitionJoinInfo._read_from_metadata_file(metadata_file)
        else:
            raise FileNotFoundError(
                f"_metadata or partition join info file is required in catalog directory {catalog_base_dir}"
            )
        return cls(pixel_frame, catalog_base_dir)

    @classmethod
    def read_from_file(cls, metadata_file: str | Path | UPath, strict: bool = False) -> PartitionJoinInfo:
        """Read partition join info from a `_metadata` file to create an object

        Args:
            metadata_file (UPath): path to the `_metadata` file
            strict (bool): use strict parsing of _metadata file. this is slower, but
                gives more helpful error messages in the case of invalid data.

        Returns:
            A `PartitionJoinInfo` object with the data from the file
        """
        return cls(cls._read_from_metadata_file(metadata_file, strict))

    @classmethod
    def _read_from_metadata_file(
        cls, metadata_file: str | Path | UPath, strict: bool = False
    ) -> pd.DataFrame:
        """Read partition join info from a `_metadata` file to create an object

        Args:
            metadata_file (UPath): path to the `_metadata` file
            strict (bool): use strict parsing of _metadata file. this is slower, but
                gives more helpful error messages in the case of invalid data.

        Returns:
            A `PartitionJoinInfo` object with the data from the file
        """
        if strict:
            pixel_frame = pd.DataFrame(
                [
                    (
                        row_group_stat_single_value(row_group, cls.PRIMARY_ORDER_COLUMN_NAME),
                        row_group_stat_single_value(row_group, cls.PRIMARY_PIXEL_COLUMN_NAME),
                        row_group_stat_single_value(row_group, cls.JOIN_ORDER_COLUMN_NAME),
                        row_group_stat_single_value(row_group, cls.JOIN_PIXEL_COLUMN_NAME),
                    )
                    for row_group in read_row_group_fragments(metadata_file)
                ],
                columns=cls.COLUMN_NAMES,
            )
        else:
            total_metadata = file_io.read_parquet_metadata(metadata_file)
            num_row_groups = total_metadata.num_row_groups

            first_row_group = total_metadata.row_group(0)
            norder_column = -1
            npix_column = -1
            join_norder_column = -1
            join_npix_column = -1

            for i in range(0, first_row_group.num_columns):
                column = first_row_group.column(i)
                if column.path_in_schema == cls.PRIMARY_ORDER_COLUMN_NAME:
                    norder_column = i
                elif column.path_in_schema == cls.PRIMARY_PIXEL_COLUMN_NAME:
                    npix_column = i
                elif column.path_in_schema == cls.JOIN_ORDER_COLUMN_NAME:
                    join_norder_column = i
                elif column.path_in_schema == cls.JOIN_PIXEL_COLUMN_NAME:
                    join_npix_column = i

            missing_columns = []
            if norder_column == -1:
                missing_columns.append(cls.PRIMARY_ORDER_COLUMN_NAME)
            if npix_column == -1:
                missing_columns.append(cls.PRIMARY_PIXEL_COLUMN_NAME)
            if join_norder_column == -1:
                missing_columns.append(cls.JOIN_ORDER_COLUMN_NAME)
            if join_npix_column == -1:
                missing_columns.append(cls.JOIN_PIXEL_COLUMN_NAME)

            if len(missing_columns) != 0:
                raise ValueError(f"Metadata missing columns ({missing_columns})")

            row_group_index = np.arange(0, num_row_groups)
            pixel_frame = pd.DataFrame(
                [
                    (
                        total_metadata.row_group(index).column(norder_column).statistics.min,
                        total_metadata.row_group(index).column(npix_column).statistics.min,
                        total_metadata.row_group(index).column(join_norder_column).statistics.min,
                        total_metadata.row_group(index).column(join_npix_column).statistics.min,
                    )
                    for index in row_group_index
                ],
                columns=cls.COLUMN_NAMES,
            )

        return pixel_frame

    @classmethod
    def read_from_csv(cls, partition_join_info_file: str | Path | UPath) -> PartitionJoinInfo:
        """Read partition join info from a `partition_join_info.csv` file to create an object

        Args:
            partition_join_info_file (UPath): path to the `partition_join_info.csv` file

        Returns:
            A `PartitionJoinInfo` object with the data from the file
        """
        return cls(cls._read_from_csv(partition_join_info_file))

    @classmethod
    def _read_from_csv(cls, partition_join_info_file: str | Path | UPath) -> pd.DataFrame:
        """Read partition join info from a `partition_join_info.csv` file to create an object

        Args:
            partition_join_info_file (UPath): path to the `partition_join_info.csv` file

        Returns:
            A `PartitionJoinInfo` object with the data from the file
        """
        if not file_io.does_file_or_directory_exist(partition_join_info_file):
            raise FileNotFoundError(
                f"No partition join info found where expected: {str(partition_join_info_file)}"
            )

        return file_io.load_csv_to_pandas(partition_join_info_file)



"""Catalog data wrappers"""

from .association_catalog import AssociationCatalog
from .association_catalog.partition_join_info import PartitionJoinInfo
from .catalog import Catalog
from .catalog_type import CatalogType
from .dataset.dataset import Dataset
from .dataset.table_properties import TableProperties
from .margin_cache.margin_catalog import MarginCatalog
from .partition_info import PartitionInfo



import os

import pyarrow as pa
import pytest

from hats.catalog import CatalogType, MarginCatalog, PartitionInfo, TableProperties
from hats.loaders import read_hats
from hats.pixel_math import HealpixPixel


def test_init_catalog(margin_catalog_info, margin_catalog_pixels):
    catalog = MarginCatalog(margin_catalog_info, margin_catalog_pixels)
    assert catalog.catalog_name == margin_catalog_info.catalog_name
    assert catalog.get_healpix_pixels() == margin_catalog_pixels
    assert catalog.catalog_info == margin_catalog_info

    assert len(catalog.get_healpix_pixels()) == len(margin_catalog_pixels)
    for hp_pixel in catalog.get_healpix_pixels():
        assert hp_pixel in margin_catalog_pixels
        assert hp_pixel in catalog.pixel_tree


def test_read_from_file(margin_catalog_path, margin_catalog_pixels, margin_catalog_schema):
    catalog = read_hats(margin_catalog_path)

    assert isinstance(catalog, MarginCatalog)
    assert catalog.on_disk
    assert catalog.catalog_path == margin_catalog_path
    assert len(catalog.get_healpix_pixels()) == len(margin_catalog_pixels)
    assert catalog.get_healpix_pixels() == margin_catalog_pixels

    info = catalog.catalog_info
    assert info.catalog_name == "small_sky_order1_margin"
    assert info.catalog_type == CatalogType.MARGIN
    assert info.ra_column == "ra"
    assert info.dec_column == "dec"
    assert info.primary_catalog == "small_sky_order1"
    assert info.margin_threshold == 7200

    assert isinstance(catalog.schema, pa.Schema)
    assert catalog.schema.equals(margin_catalog_schema)


# pylint: disable=duplicate-code
def test_empty_directory(tmp_path, margin_cache_catalog_info_data, margin_catalog_pixels):
    """Test loading empty or incomplete data"""
    ## Path doesn't exist
    with pytest.raises(FileNotFoundError):
        read_hats(os.path.join("path", "empty"))

    catalog_path = tmp_path / "empty"
    os.makedirs(catalog_path, exist_ok=True)

    ## Path exists but there's nothing there
    with pytest.raises(FileNotFoundError, match="properties file"):
        read_hats(catalog_path)

    ## catalog_info file exists - getting closer
    properties = TableProperties(**margin_cache_catalog_info_data)
    properties.to_properties_file(catalog_path)

    with pytest.raises(FileNotFoundError, match="metadata"):
        read_hats(catalog_path)

    ## Now we create the needed _metadata and everything is right.
    part_info = PartitionInfo.from_healpix(margin_catalog_pixels)
    part_info.write_to_metadata_files(catalog_path=catalog_path)

    with pytest.warns(UserWarning, match="slow"):
        catalog = read_hats(catalog_path)
    assert catalog.catalog_name == margin_cache_catalog_info_data["catalog_name"]


def test_margin_filter(margin_catalog_info, margin_catalog_pixels):
    catalog = MarginCatalog(margin_catalog_info, margin_catalog_pixels)
    pixels = [HealpixPixel(1, 44)]
    filtered_catalog = catalog.filter_from_pixel_list(pixels)
    assert filtered_catalog.get_healpix_pixels() == [
        HealpixPixel(1, 44),
        HealpixPixel(1, 45),
        HealpixPixel(1, 46),
        HealpixPixel(1, 47),
    ]


def test_margin_filter_invalid_size(margin_catalog_info, margin_catalog_pixels):
    margin_catalog_info.margin_threshold = 10000000
    catalog = MarginCatalog(margin_catalog_info, margin_catalog_pixels)
    pixels = [HealpixPixel(1, 44)]
    with pytest.raises(ValueError, match="greater than the size of a pixel"):
        catalog.filter_from_pixel_list(pixels)



"""Generate a molleview map with the pixel densities of the catalog

NB: Testing validity of generated plots is currently not tested in our unit test suite.
"""

from typing import List

import healpy as hp
import matplotlib.colors as mcolors
import numpy as np
from matplotlib import pyplot as plt

import hats.pixel_math.healpix_shim as hs
from hats.catalog import Catalog
from hats.io import file_io, paths
from hats.pixel_math import HealpixPixel


def _read_point_map(catalog_base_dir):
    """Read the object spatial distribution information from a healpix FITS file.

    Args:
        catalog_base_dir: path to a catalog
    Returns:
        one-dimensional numpy array of long integers where the value at each index
        corresponds to the number of objects found at the healpix pixel.
    """
    map_file_pointer = paths.get_point_map_file_pointer(catalog_base_dir)
    return file_io.read_fits_image(map_file_pointer)


def plot_points(catalog: Catalog, projection="moll", **kwargs):
    """Create a visual map of the input points of an in-memory catalog.

    Args:
        catalog (`hats.catalog.Catalog`) Catalog to display
        projection (str) The map projection to use. Valid values include:
            - moll - Molleweide projection (default)
            - gnom - Gnomonic projection
            - cart - Cartesian projection
            - orth - Orthographic projection
    """
    if not catalog.on_disk:
        raise ValueError("on disk catalog required for point-wise visualization")
    point_map = _read_point_map(catalog.catalog_base_dir)
    _plot_healpix_map(point_map, projection, f"Catalog point density map - {catalog.catalog_name}", **kwargs)


def plot_pixels(catalog: Catalog, projection="moll", **kwargs):
    """Create a visual map of the pixel density of the catalog.

    Args:
        catalog (`hats.catalog.Catalog`) Catalog to display
        projection (str) The map projection to use. Valid values include:
            - moll - Molleweide projection (default)
            - gnom - Gnomonic projection
            - cart - Cartesian projection
            - orth - Orthographic projection
    """
    pixels = catalog.get_healpix_pixels()
    plot_pixel_list(
        pixels=pixels,
        plot_title=f"Catalog pixel density map - {catalog.catalog_name}",
        projection=projection,
        **kwargs,
    )


def plot_pixel_list(pixels: List[HealpixPixel], plot_title: str = "", projection="moll", **kwargs):
    """Create a visual map of the pixel density of a list of pixels.

    Args:
        pixels: healpix pixels (order and pixel number) to visualize
        plot_title (str): heading for the plot
        projection (str) The map projection to use. Valid values include:
            - moll - Molleweide projection (default)
            - gnom - Gnomonic projection
            - cart - Cartesian projection
            - orth - Orthographic projection
    """
    max_order = np.max(pixels).order
    min_order = np.min(pixels).order

    if max_order == min_order:
        colors = [plt.cm.viridis(0.0), plt.cm.viridis(0.1)]  # pylint: disable=no-member
        cmap = mcolors.LinearSegmentedColormap.from_list("my_colormap", colors, 1)
        kwargs["cbar"] = False
        plot_title = f"Norder {max_order} {plot_title}"
    else:
        num_colors = max_order - min_order + 1
        colors = plt.cm.viridis(np.linspace(0, 1, num_colors))  # pylint: disable=no-member
        cmap = mcolors.LinearSegmentedColormap.from_list("my_colormap", colors, num_colors)

    order_map = np.full(hs.order2npix(max_order), hs.unseen_pixel())

    for pixel in pixels:
        explosion_factor = 4 ** (max_order - pixel.order)
        exploded_pixels = [
            *range(
                pixel.pixel * explosion_factor,
                (pixel.pixel + 1) * explosion_factor,
            )
        ]
        order_map[exploded_pixels] = pixel.order
    _plot_healpix_map(order_map, projection, plot_title, cmap=cmap, **kwargs)


def get_projection_method(projection):
    """Get the healpy plotting method for a specified projection string

    Args:
        projection (str):  The map projection to use. Valid values include:
            - moll - Molleweide projection (default)
            - gnom - Gnomonic projection
            - cart - Cartesian projection
            - orth - Orthographic projection

    Returns:
        The healpy method that plots a HEALPix map with the specified projection
    """
    if projection == "moll":
        projection_method = hp.mollview
    elif projection == "gnom":
        projection_method = hp.gnomview
    elif projection == "cart":
        projection_method = hp.cartview
    elif projection == "orth":
        projection_method = hp.orthview
    else:
        raise NotImplementedError(f"unknown projection: {projection}")
    return projection_method


def _plot_healpix_map(healpix_map, projection, title, cmap="viridis", **kwargs):
    """Perform the plotting of a healpix pixel map.

    Args:
        healpix_map: array containing the map
        projection: projection type to display
        title: title used in image plot
        cmap: matplotlib colormap to use
    """
    projection_method = get_projection_method(projection)

    projection_method(healpix_map, title=title, nest=True, cmap=cmap, **kwargs)
    plt.plot()



from __future__ import annotations

from enum import Enum
from typing import List, Tuple

import numpy as np


class ValidatorsErrors(str, Enum):
    """Error messages for the coordinate validators"""

    INVALID_DEC = "declination must be in the -90.0 to 90.0 degree range"
    INVALID_RADIUS = "cone radius must be positive"
    INVALID_NUM_VERTICES = "polygon must contain a minimum of 3 vertices"
    DUPLICATE_VERTICES = "polygon has duplicated vertices"
    DEGENERATE_POLYGON = "polygon is degenerate"
    INVALID_RADEC_RANGE = "invalid ra or dec range"


def validate_radius(radius_arcsec: float):
    """Validates that a cone search radius is positive

    Args:
        radius_arcsec (float): The cone radius, in arcseconds

    Raises:
        ValueError: if radius is non-positive
    """
    if radius_arcsec <= 0:
        raise ValueError(ValidatorsErrors.INVALID_RADIUS.value)


def validate_declination_values(dec: float | List[float]):
    """Validates that declination values are in the [-90,90] degree range

    Args:
        dec (float | List[float]): The declination values to be validated

    Raises:
        ValueError: if declination values are not in the [-90,90] degree range
    """
    dec_values = np.array(dec)
    lower_bound, upper_bound = -90.0, 90.0
    if not np.all((dec_values >= lower_bound) & (dec_values <= upper_bound)):
        raise ValueError(ValidatorsErrors.INVALID_DEC.value)


def validate_polygon(vertices: np.ndarray):
    """Checks if the polygon contain a minimum of three vertices, that they are
    unique and that the polygon does not fall on a great circle.

    Args:
        vertices (np.ndarray): The polygon vertices, in cartesian coordinates

    Raises:
        ValueError: exception if the polygon is invalid.
    """
    if len(vertices) < 3:
        raise ValueError(ValidatorsErrors.INVALID_NUM_VERTICES.value)
    if len(vertices) != len(np.unique(vertices, axis=0)):
        raise ValueError(ValidatorsErrors.DUPLICATE_VERTICES.value)
    if is_polygon_degenerate(vertices):
        raise ValueError(ValidatorsErrors.DEGENERATE_POLYGON.value)


def is_polygon_degenerate(vertices: np.ndarray) -> bool:
    """Checks if all the vertices of the polygon are contained in a same plane.
    If the plane intersects the center of the sphere, the polygon is degenerate.

    Args:
        vertices (np.ndarray): The polygon vertices, in cartesian coordinates

    Returns:
        A boolean, which is True if the polygon is degenerate, i.e. if it falls
        on a great circle, False otherwise.
    """
    # Calculate the normal vector of the plane using three of the vertices
    normal_vector = np.cross(vertices[1] - vertices[0], vertices[2] - vertices[0])

    # Check if the other vertices lie on the same plane
    for vertex in vertices[3:]:
        dot_product = np.dot(normal_vector, vertex - vertices[0])
        if not np.isclose(dot_product, 0):
            return False

    # Check if the plane intersects the sphere's center. If it does,
    # the polygon is degenerate and therefore, invalid.
    center_distance = np.dot(normal_vector, vertices[0])
    return bool(np.isclose(center_distance, 0))


def validate_box_search(ra: Tuple[float, float] | None, dec: Tuple[float, float] | None):
    """Checks if ra and dec values are valid for the box search.

    - At least one range of ra or dec must have been provided
    - Ranges must be pairs of non-duplicate minimum and maximum values, in degrees
    - Declination values, if existing, must be in ascending order
    - Declination values, if existing, must be in the [-90,90] degree range

    Args:
        ra (Tuple[float, float]): Right ascension range, in degrees
        dec (Tuple[float, float]): Declination range, in degrees
    """
    invalid_range = False
    if ra is not None:
        if len(ra) != 2 or len(ra) != len(set(ra)):
            invalid_range = True
    if dec is not None:
        if len(dec) != 2 or dec[0] >= dec[1]:
            invalid_range = True
        validate_declination_values(list(dec))
    if (ra is None and dec is None) or invalid_range:
        raise ValueError(ValidatorsErrors.INVALID_RADEC_RANGE.value)



import numpy as np
from numpy.testing import assert_allclose, assert_array_equal

from hats.pixel_math import healpix_shim as hps


def test_avgsize2mindist2avgsize():
    """Test that avgsize2mindist is inverse of mindist2avgsize"""
    avgsize = np.logspace(-5, 5, 21)
    assert_allclose(hps.mindist2avgsize(hps.avgsize2mindist(avgsize)), avgsize, rtol=1e-12)


def test_mindist2avgsize2mindist():
    """Test that mindist2avgsize is inverse of avgsize2mindist"""
    mindist = np.logspace(-3.2, 2.8, 21)
    assert_allclose(hps.avgsize2mindist(hps.mindist2avgsize(mindist)), mindist, rtol=1e-12)


def test_order2avgsize2order():
    """Test that avgsize2order is inverse of hps.nside2resol(hps.order2nside, arcmin=True)"""
    order = np.arange(20)
    nside = hps.order2nside(order)
    assert_array_equal(hps.avgsize2order(hps.nside2resol(nside, arcmin=True)), order)


def test_margin2order():
    """Test margin2order for some pre-computed values"""
    margin_thr_arcmin = np.array([1 / 60, 10 / 60, 1, 5, 60])
    orders = np.array([17, 13, 11, 8, 5])
    assert_array_equal(hps.margin2order(margin_thr_arcmin), orders)



import os

import pandas as pd
import pyarrow as pa
import pytest

from hats.catalog.association_catalog.association_catalog import AssociationCatalog
from hats.catalog.association_catalog.partition_join_info import PartitionJoinInfo
from hats.catalog.dataset.table_properties import TableProperties
from hats.loaders import read_hats
from hats.pixel_math import HealpixPixel


def test_init_catalog(association_catalog_info, association_catalog_join_pixels):
    catalog = AssociationCatalog(
        association_catalog_info, [HealpixPixel(0, 11)], association_catalog_join_pixels
    )
    assert catalog.catalog_name == association_catalog_info.catalog_name
    pd.testing.assert_frame_equal(catalog.get_join_pixels(), association_catalog_join_pixels)
    assert catalog.get_healpix_pixels() == [HealpixPixel(0, 11)]
    assert catalog.catalog_info == association_catalog_info

    assert len(catalog.get_healpix_pixels()) == len([HealpixPixel(0, 11)])
    for hp_pixel in catalog.get_healpix_pixels():
        assert hp_pixel in [HealpixPixel(0, 11)]
        assert hp_pixel in catalog.pixel_tree


def test_wrong_join_pixels_type(association_catalog_info):
    with pytest.raises(TypeError, match="join_pixels"):
        AssociationCatalog(association_catalog_info, [HealpixPixel(0, 11)], "test")


def test_different_join_pixels_type(association_catalog_info, association_catalog_join_pixels):
    partition_join_info = PartitionJoinInfo(association_catalog_join_pixels)
    catalog = AssociationCatalog(association_catalog_info, [HealpixPixel(0, 11)], partition_join_info)
    pd.testing.assert_frame_equal(catalog.get_join_pixels(), association_catalog_join_pixels)


def test_read_from_file(
    association_catalog_path, association_catalog_join_pixels, association_catalog_schema
):
    catalog = read_hats(association_catalog_path)

    assert isinstance(catalog, AssociationCatalog)
    assert catalog.on_disk
    assert catalog.catalog_path == association_catalog_path
    assert len(catalog.get_join_pixels()) == 4
    assert len(catalog.get_healpix_pixels()) == 1
    pd.testing.assert_frame_equal(catalog.get_join_pixels(), association_catalog_join_pixels)
    assert catalog.get_healpix_pixels() == [HealpixPixel(0, 11)]

    info = catalog.catalog_info
    assert info.primary_catalog == "small_sky"
    assert info.primary_column == "id"
    assert info.join_catalog == "small_sky_order1"
    assert info.join_column == "id"

    assert isinstance(catalog.schema, pa.Schema)
    assert catalog.schema.equals(association_catalog_schema)


def test_empty_directory(tmp_path, association_catalog_info_data, association_catalog_join_pixels):
    """Test loading empty or incomplete data"""
    ## Path doesn't exist
    with pytest.raises(FileNotFoundError):
        read_hats(os.path.join("path", "empty"))

    catalog_path = tmp_path / "empty"
    os.makedirs(catalog_path, exist_ok=True)

    ## Path exists but there's nothing there
    with pytest.raises(FileNotFoundError, match="properties file"):
        AssociationCatalog.read_hats(catalog_path)

    ## catalog_info file exists - getting closer
    properties = TableProperties(**association_catalog_info_data)
    properties.to_properties_file(catalog_path)

    with pytest.raises(FileNotFoundError, match="metadata"):
        read_hats(catalog_path)

    ## Now we create the needed _metadata and everything is right.
    part_info = PartitionJoinInfo(association_catalog_join_pixels)
    part_info.write_to_metadata_files(catalog_path=catalog_path)
    with pytest.warns(UserWarning, match="slow"):
        catalog = read_hats(catalog_path)
    assert catalog.catalog_name == association_catalog_info_data["catalog_name"]


def test_csv_round_trip(tmp_path, association_catalog_info_data, association_catalog_join_pixels):
    ## Path doesn't exist
    with pytest.raises(FileNotFoundError):
        read_hats(os.path.join("path", "empty"))

    catalog_path = tmp_path / "empty"
    os.makedirs(catalog_path, exist_ok=True)

    properties = TableProperties(**association_catalog_info_data)
    properties.to_properties_file(catalog_path)

    with pytest.raises(FileNotFoundError, match="partition"):
        read_hats(catalog_path)

    file_name = catalog_path / "partition_info.csv"
    with open(file_name, "w", encoding="utf-8") as metadata_file:
        # dump some garbage in there - just needs to exist.
        metadata_file.write("Norder,Npix")
    with pytest.raises(FileNotFoundError, match="partition"):
        read_hats(catalog_path)

    part_info = PartitionJoinInfo(association_catalog_join_pixels)
    part_info.write_to_csv(catalog_path=catalog_path)

    with pytest.warns(UserWarning, match="_common_metadata or _metadata files not found"):
        catalog = read_hats(catalog_path)
    pd.testing.assert_frame_equal(catalog.get_join_pixels(), association_catalog_join_pixels)



from __future__ import annotations

from pathlib import Path
from typing import Tuple, Union

import pandas as pd
import pyarrow as pa
from mocpy import MOC
from upath import UPath

from hats.catalog.association_catalog.partition_join_info import PartitionJoinInfo
from hats.catalog.dataset.table_properties import TableProperties
from hats.catalog.healpix_dataset.healpix_dataset import HealpixDataset, PixelInputTypes
from hats.io import file_io, paths


class AssociationCatalog(HealpixDataset):
    """A HATS Catalog for enabling fast joins between two HATS catalogs

    Catalogs of this type are partitioned based on the partitioning of the left catalog.
    The `partition_join_info` metadata file specifies all pairs of pixels in the Association
    Catalog, corresponding to each pair of partitions in each catalog that contain rows to join.
    """

    JoinPixelInputTypes = Union[list, pd.DataFrame, PartitionJoinInfo]

    def __init__(
        self,
        catalog_info: TableProperties,
        pixels: PixelInputTypes,
        join_pixels: JoinPixelInputTypes,
        catalog_path=None,
        moc: MOC | None = None,
        schema: pa.Schema | None = None,
    ) -> None:
        super().__init__(catalog_info, pixels, catalog_path, moc=moc, schema=schema)
        self.join_info = self._get_partition_join_info_from_pixels(join_pixels)

    def get_join_pixels(self) -> pd.DataFrame:
        """Get join pixels listing all pairs of pixels from left and right catalogs that contain
        matching association rows

        Returns:
            pd.DataFrame with each row being a pair of pixels from the primary and join catalogs
        """
        return self.join_info.data_frame

    @staticmethod
    def _get_partition_join_info_from_pixels(
        join_pixels: JoinPixelInputTypes,
    ) -> PartitionJoinInfo:
        if isinstance(join_pixels, PartitionJoinInfo):
            return join_pixels
        if isinstance(join_pixels, pd.DataFrame):
            return PartitionJoinInfo(join_pixels)
        raise TypeError("join_pixels must be of type PartitionJoinInfo or DataFrame")

    @classmethod
    def _read_args(
        cls, catalog_base_dir: str | Path | UPath
    ) -> Tuple[TableProperties, PixelInputTypes, JoinPixelInputTypes]:  # type: ignore[override]
        args = super()._read_args(catalog_base_dir)
        partition_join_info = PartitionJoinInfo.read_from_dir(catalog_base_dir)
        return args + (partition_join_info,)

    @classmethod
    def _check_files_exist(cls, catalog_base_dir: str | Path | UPath):
        super()._check_files_exist(catalog_base_dir)
        partition_join_info_file = paths.get_partition_join_info_pointer(catalog_base_dir)
        metadata_file = paths.get_parquet_metadata_pointer(catalog_base_dir)
        if not (
            file_io.does_file_or_directory_exist(partition_join_info_file)
            or file_io.does_file_or_directory_exist(metadata_file)
        ):
            raise FileNotFoundError(
                f"_metadata or partition join info file is required in catalog directory {catalog_base_dir}"
            )



"""Methods for creating partitioned data paths"""

from __future__ import annotations

import re
from pathlib import Path
from typing import Dict, List
from urllib.parse import urlencode

from fsspec.implementations.http import HTTPFileSystem
from upath import UPath

from hats.io.file_io.file_pointer import get_upath
from hats.pixel_math.healpix_pixel import INVALID_PIXEL, HealpixPixel

PARTITION_ORDER = "Norder"
PARTITION_DIR = "Dir"
PARTITION_PIXEL = "Npix"

MARGIN_ORDER = "margin_Norder"
MARGIN_DIR = "margin_Dir"
MARGIN_PIXEL = "margin_Npix"

DATASET_DIR = "dataset"
PARTITION_INFO_FILENAME = "partition_info.csv"
PARTITION_JOIN_INFO_FILENAME = "partition_join_info.csv"
PARQUET_METADATA_FILENAME = "_metadata"
PARQUET_COMMON_METADATA_FILENAME = "_common_metadata"
POINT_MAP_FILENAME = "point_map.fits"


def pixel_directory(
    catalog_base_dir: str | Path | UPath | None,
    pixel_order: int,
    pixel_number: int | None = None,
    directory_number: int | None = None,
) -> UPath:
    """Create path pointer for a pixel directory. This will not create the directory.

    One of pixel_number or directory_number is required. The directory name will
    take the HiPS standard form of::

        <catalog_base_dir>/dataset/Norder=<pixel_order>/Dir=<directory number>

    Where the directory number is calculated using integer division as::

        (pixel_number/10000)*10000

    Args:
        catalog_base_dir (UPath): base directory of the catalog (includes catalog name)
        pixel_order (int): the healpix order of the pixel
        directory_number (int): directory number
        pixel_number (int): the healpix pixel
    Returns:
        UPath directory name
    """
    norder = int(pixel_order)
    if directory_number is not None:
        ndir = directory_number
    elif pixel_number is not None:
        npix = int(pixel_number)
        ndir = int(npix / 10_000) * 10_000
    else:
        raise ValueError("One of pixel_number or directory_number is required to create pixel directory")

    return (
        get_upath(catalog_base_dir) / DATASET_DIR / f"{PARTITION_ORDER}={norder}" / f"{PARTITION_DIR}={ndir}"
    )


def get_healpix_from_path(path: str) -> HealpixPixel:
    """Find the `pixel_order` and `pixel_number` from a string like the
    following::

        Norder=<pixel_order>/Dir=<directory number>/Npix=<pixel_number>.parquet

    NB: This expects the format generated by the `pixel_catalog_file` method

    Args:
        path (str): path to parse

    Returns:
        Constructed HealpixPixel object representing the pixel in the path.
        `INVALID_PIXEL` if the path doesn't match the expected pattern for any reason.
    """
    match = re.match(r".*Norder=(\d*).*Npix=(\d*).*", path)
    if not match:
        return INVALID_PIXEL
    order, pixel = match.groups()
    return HealpixPixel(int(order), int(pixel))


def pixel_catalog_files(
    catalog_base_dir: str | Path | UPath | None,
    pixels: List[HealpixPixel],
    query_params: Dict | None = None,
) -> List[UPath]:
    """Create a list of path *pointers* for pixel catalog files. This will not create the directory
    or files.

    The catalog file names will take the HiPS standard form of::

        <catalog_base_dir>/Norder=<pixel_order>/Dir=<directory number>/Npix=<pixel_number>.parquet

    Where the directory number is calculated using integer division as::

        (pixel_number/10000)*10000

    Args:
        catalog_base_dir (UPath): base directory of the catalog (includes catalog name)
        pixels (List[HealpixPixel]): the healpix pixels to create pointers to
        query_params (dict): Params to append to URL. Ex: {'cols': ['ra', 'dec'], 'fltrs': ['r>=10', 'g<18']}

    Returns (List[str]):
        A list of paths to the pixels, in the same order as the input pixel list.
    """
    catalog_base_dir = get_upath(catalog_base_dir) / DATASET_DIR
    fs = catalog_base_dir.fs
    base_path = str(catalog_base_dir)
    if not base_path.endswith(fs.sep):
        base_path += fs.sep

    url_params = ""
    if isinstance(fs, HTTPFileSystem) and query_params:
        url_params = dict_to_query_urlparams(query_params)

    return [
        base_path
        + fs.sep.join(
            [
                f"{PARTITION_ORDER}={pixel.order}",
                f"{PARTITION_DIR}={pixel.dir}",
                f"{PARTITION_PIXEL}={pixel.pixel}.parquet" + url_params,
            ]
        )
        for pixel in pixels
    ]


def dict_to_query_urlparams(query_params: Dict | None = None) -> str:
    """Converts a dictionary to a url query parameter string

    Args:
        query_params (dict): dictionary of query parameters.
    Returns:
        query parameter string to append to a url
    """

    if not query_params:
        return ""

    query = {}
    for key, value in query_params.items():
        if not all([key, value]):
            continue
        if isinstance(value, list):
            value = ",".join(value).replace(" ", "")
        query[key] = value

    if not query:
        return ""

    # Build the query string and add the "?" prefix
    url_params = "?" + urlencode(query, doseq=True)
    return url_params


def pixel_catalog_file(
    catalog_base_dir: str | Path | UPath | None, pixel: HealpixPixel, query_params: Dict | None = None
) -> UPath:
    """Create path *pointer* for a pixel catalog file. This will not create the directory
    or file.

    The catalog file name will take the HiPS standard form of::

        <catalog_base_dir>/Norder=<pixel_order>/Dir=<directory number>/Npix=<pixel_number>.parquet

    Where the directory number is calculated using integer division as::

        (pixel_number/10000)*10000

    Args:
        catalog_base_dir (UPath): base directory of the catalog (includes catalog name)
        pixel (HealpixPixel): the healpix pixel to create path to
        query_params (dict): Params to append to URL. Ex: {'cols': ['ra', 'dec'], 'fltrs': ['r>=10', 'g<18']}
    Returns:
        string catalog file name
    """
    catalog_base_dir = get_upath(catalog_base_dir)

    url_params = ""
    if isinstance(catalog_base_dir.fs, HTTPFileSystem) and query_params:
        url_params = dict_to_query_urlparams(query_params)

    return (
        catalog_base_dir
        / DATASET_DIR
        / f"{PARTITION_ORDER}={pixel.order}"
        / f"{PARTITION_DIR}={pixel.dir}"
        / f"{PARTITION_PIXEL}={pixel.pixel}.parquet{url_params}"
    )


def get_partition_info_pointer(catalog_base_dir: str | Path | UPath) -> UPath:
    """Get file pointer to `partition_info.csv` metadata file

    Args:
        catalog_base_dir: pointer to base catalog directory
    Returns:
        File Pointer to the catalog's `partition_info.csv` file
    """
    return get_upath(catalog_base_dir) / PARTITION_INFO_FILENAME


def get_common_metadata_pointer(catalog_base_dir: str | Path | UPath) -> UPath:
    """Get file pointer to `_common_metadata` parquet metadata file

    Args:
        catalog_base_dir: pointer to base catalog directory
    Returns:
        File Pointer to the catalog's `_common_metadata` file
    """
    return get_upath(catalog_base_dir) / DATASET_DIR / PARQUET_COMMON_METADATA_FILENAME


def get_parquet_metadata_pointer(catalog_base_dir: str | Path | UPath) -> UPath:
    """Get file pointer to `_metadata` parquet metadata file

    Args:
        catalog_base_dir: pointer to base catalog directory
    Returns:
        File Pointer to the catalog's `_metadata` file
    """
    return get_upath(catalog_base_dir) / DATASET_DIR / PARQUET_METADATA_FILENAME


def get_point_map_file_pointer(catalog_base_dir: str | Path | UPath) -> UPath:
    """Get file pointer to `point_map.fits` FITS image file.

    Args:
        catalog_base_dir: pointer to base catalog directory
    Returns:
        File Pointer to the catalog's `point_map.fits` FITS image file.
    """
    return get_upath(catalog_base_dir) / POINT_MAP_FILENAME


def get_partition_join_info_pointer(catalog_base_dir: str | Path | UPath) -> UPath:
    """Get file pointer to `partition_join_info.csv` association metadata file

    Args:
        catalog_base_dir: pointer to base catalog directory
    Returns:
        File Pointer to the catalog's `partition_join_info.csv` association metadata file
    """
    return get_upath(catalog_base_dir) / PARTITION_JOIN_INFO_FILENAME
